import logging
from pathlib import Path
import duckdb
import re
# from dabbler.lsp.server_classes import InlineSqlLangServer
from lsprotocol.types import (
    CompletionItem,
    CompletionItemKind,
    CompletionItemLabelDetails
)

duckdb_types = [CompletionItem(label=f'{x[0]}') for x in 
                duckdb.execute("select distinct logical_type from duckdb_types()").fetchall()]


duckdb_extensions = [CompletionItem(label=f"'{x[0]}'") for x in 
                duckdb.execute("select distinct extension_name from duckdb_extensions()").fetchall()]


duckdb_settings = [CompletionItem(label=f'{x[0]}') for x in 
                duckdb.execute("select distinct name from duckdb_settings()").fetchall()]

duckdb_kw_comp = [CompletionItem(label=f'{x[0]}',sort_text="99") for x in 
                duckdb.execute("select upper(keyword_name) from duckdb_keywords() where keyword_category = 'reserved'").fetchall()]

duckdb_pragmas = [CompletionItem(label=f'{x[0]}',sort_text="99") for x in 
                duckdb.execute("select distinct function_name from duckdb_functions() where function_type ='pragma'").fetchall()]

types_to_exclude = set(['.py','.ipynb'])
file_path_completion_pattern = r'''.*('./)([A-Za-z0-9./-_, ]+)?$'''
file_path_completion_regex = re.compile(file_path_completion_pattern)


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


class PathCompleter:
    def __init__(self,cwd:str,search_path:str,logger:logging.Logger,types_to_exclude=types_to_exclude) -> None:
        self.cwd = None
        self.search_path = None
        self.log = logger
        if cwd and Path(cwd).is_dir():
            self.cwd = Path(cwd)
        if search_path and Path(search_path).is_dir():
            self.search_path = Path(search_path)
            
    def get_items(self, search:str):
        path = None
        file = None
        if search is None:
            search = ''
        
        if '/' in search:
            path,file = search.rsplit('/',1)
        
        if path is None:
            path = ''
        
        if file is None:
            file = ''
        
        items:list[Path] = []

        if self.cwd:
            items.extend([f for f in self.cwd.joinpath(path).glob(f'{file}*') if f.suffix.lower() not in types_to_exclude])
        
        file_names = {x.name for x in items}
        
        if self.search_path:
            glob = self.search_path.joinpath(path).glob(f'{file}*')
            files = [f for f in glob if f.name not in file_names and f.suffix.lower() not in types_to_exclude]
            items.extend(files)
        
        
        comp_items = [CompletionItem(
            label=f.name,
            sort_text="1",
            kind=CompletionItemKind.Folder if f.is_dir() else CompletionItemKind.File,
            label_details=CompletionItemLabelDetails(
                detail=None,
                description=sizeof_fmt(f.stat().st_size) if f.is_file() else None,
            )
            
            )
                      
                      for f in items]
        
        self.log.debug([search,path,file,items])
        return comp_items
    

def pathlib_completetions(text:str, path_dict:dict[str,str],logger):
    if len(path_dict) == 0:
        return
    path_strings = '|'.join([x.replace('.',r'\.') for x in path_dict.keys()])
    slash_pattern = r""".*(^|\s+|\(|=)(?P<path_ojb>(""" + path_strings + r"""))(?P<parent>(\.parent|\.joinpath\(['"][A-Za-z0-9./-_, ]+['"]\))+)?(\s+)?/(\s+)?['"](?P<search>[A-Za-z0-9./-_, ]+)?$"""
    join__pattern = r""".*(^|\s+|\(|=)(?P<path_ojb>(""" + path_strings + r"""))(?P<parent>(\.parent|\.joinpath\(['"][A-Za-z0-9./-_, ]+['"]\))+)?\.joinpath\(["'](?P<search>[A-Za-z0-9./-_, ]+)?$"""

    m = re.search(slash_pattern,text)
    if not m:
        m = re.search(join__pattern,text)
    if not m:
        return
    
    groups = m.groupdict()
    if not groups['path_ojb']:
        return
    path_ojb = groups['path_ojb']
    if path_ojb not in path_dict:
        return
    _path = Path(path_dict[path_ojb])
    if groups['parent']:
        _path = eval(f'_path{groups["parent"]}')
    if not _path.is_dir():
        return
    
    if groups['search']:
        search = groups['search']
    else:
        search = ''
    
    completer = PathCompleter(_path,None,logger)
    return completer.get_items(search)



from queue import Queue
from dabbler.common import FromLangServer, ToLangServer, KeyFile
import logging
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Slot, Signal
from IPython import get_ipython
import pickle
import time
import zmq
import duckdb
from dabbler.db_stuff import get_db_data_new

margins = 0

class Shortcut:
    def __init__(self, Seq, Parent, target):
        if not isinstance(Seq, list):
            Seq = [Seq]
        for i in Seq:
            sc = QtGui.QShortcut(QtGui.QKeySequence(i), Parent)
            sc.activated.connect(target)


class Hbox(QtWidgets.QHBoxLayout):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSpacing(margins)
        self.setContentsMargins(margins, margins, margins, margins)


class Vbox(QtWidgets.QVBoxLayout):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSpacing(margins)
        self.setContentsMargins(margins, margins, margins, margins)


class TreeItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, *args, item_type=None):
        super().__init__(*args)
        self.item_type = item_type


class IPythonExecutionMonitor(object):
    def __init__(self, ip, parent: "ZmqServer", timer: QtCore.QTimer = None):
        self.shell = ip
        self.parent = parent
        self.timer = timer

    def pre_run_cell(self, info):
        self.timer.start(1_000_000)

    def post_run_cell(self, result):
        if result.info.cell_id:
            self.timer.start(200)


class RepChannel(QtCore.QThread):
    trigger = Signal(object)

    def __init__(self, socket: zmq.Socket, log: logging.Logger = None) -> None:
        super().__init__()
        self.thead_active = True
        self.socket = socket
        self.logger = log

    def run(self):
        while self.thead_active:
            if self.socket.poll(500):
                buff = self.socket.recv()
                msg = pickle.loads(buff)
                self.trigger.emit(msg)

    def stop(self):
        self.thead_active = False
        self.quit()
        self.wait()
        self.logger.debug("rep thread stopped")
        self.finished.emit()
        self.deleteLater()


class HBChannel(QtCore.QThread):
    trigger = Signal(object)

    def __init__(self, socket: zmq.Socket, log: logging.Logger = None) -> None:
        super().__init__()
        self.thead_active = True
        self.socket = socket
        self.logger = log
        self.log("hb thread init")

    def log(self, msg):
        if self.logger:
            self.logger.debug(msg)

    def run(self):
        while self.thead_active:
            self.send_msg({"cmd": "heartbeat"})
            if not self.socket.poll(5000):
                self.log("lost connection to lang server")
                self.trigger.emit("reset_sockets")
                break
            buff = self.socket.recv()
            msg: FromLangServer = pickle.loads(buff)
            self.log(f"heartbeat channel got message {msg}")
            if msg["cmd"] != "heartbeat":
                self.log.error(f"heartbeat channel got wrong message {msg}")
                break
            time.sleep(3)
        self.quit()
        self.wait()
        self.finished.emit()
        self.deleteLater()

    def send_msg(self, msg: ToLangServer):
        buff = pickle.dumps(msg)
        self.socket.send(buff)

    def stop(self):
        self.thead_active = False
        self.quit()
        self.wait()
        self.finished.emit()
        self.deleteLater()


class ZmqServer(QtCore.QObject):
    def __init__(self, parent=None, db: duckdb.DuckDBPyConnection = None):
        super().__init__(parent)
        self.app = parent
        self.db = db.cursor()
        self.log: logging.Logger = self.app.log.getChild("gui_zmq")
        self.log.info("zmq server init")

        self.sockets_created = False
        self.connected = False
        self.connection_id: int = None
        self.new_data = False
        self.data_sent = False
        self.context = zmq.Context()
        self.db_data = get_db_data_new(self.db, self.app.file_search_path)

        self.hb_channels: list[HBChannel] = []
        self.connect_sockets()

        ip = get_ipython()
        self.ipython_cell_timer = QtCore.QTimer()
        self.ipython_cell_timer.setSingleShot(True)
        self.ipython_cell_timer.timeout.connect(self.check_for_update)
        self.monitor_ip = IPythonExecutionMonitor(ip, self, self.ipython_cell_timer)
        ip.events.register("pre_run_cell", self.monitor_ip.pre_run_cell)
        ip.events.register("post_run_cell", self.monitor_ip.post_run_cell)

    def clear_conn_info(self, delete=False):
        if delete:
            self.key_file.delete_connection(self.connection["workspace_path"])
            self.log.debug("deleted connection info from keyfile")
        self.connection = None
        self.key_file = None
        self.main_port = None
        self.handshake_port = None

    def attempt_connection(self):
        self.clear_conn_info()
        self.key_file = KeyFile()
        self.connection = self.key_file.get_connection(self.app.py_file)
        if not self.connection:
            self.log.debug("Connection info not found in keyfile")
            return False

        self.main_port = self.connection["main_port"]
        self.handshake_port = self.connection["handshake_port"]

        try:
            self.create_sockets()
            self.initialize_channels()
            return True
        except Exception as e:
            self.log.error(f"Failed to connect_: {e}")
            return False

    def connect_sockets(self):
        if self.sockets_created:
            return

        if self.attempt_connection():
            self.sockets_created = True
            self.log.debug(f"Sockets created, main port {self.main_port}, handshake port {self.handshake_port}")
            print(f'Connected to language server, workspace {self.connection["workspace_path"]}')

            if self.data_sent is False:
                self.handshake_send({"cmd": "ip_python_started", "data": 1})
                if self.app.debug:
                    self.handshake_send({"cmd": "debug", "data": True})
        else:
            self.log.debug("Failed to connect, will retry later")

    def reconnect(self):
        self.log.debug("Attempting to reconnect")
        while not self.attempt_connection():
            time.sleep(5)  # Wait for 5 seconds before trying again
        self.sockets_created = True
        self.connected = True
        self.log.debug(f"Reconnected successfully, {self.sockets_created}")
        self.send_db_data()

    def create_sockets(self):
        self.log.debug(f"connecting to main port {self.main_port}")
        self.socket = self.context.socket(zmq.PAIR)
        self.socket.bind(f"tcp://127.0.0.1:{self.main_port}")

        self.log.debug(f"connecting to handshake port {self.handshake_port}")
        self.handeshake_socket = self.context.socket(zmq.PAIR)
        self.handeshake_socket.connect(f"tcp://127.0.0.1:{self.handshake_port}")

        handshake_msg: ToLangServer = {
            "cmd": "connection_id",
            "data": self.connection["client_id"],
        }


        self.log.debug(f"sending handshake message {handshake_msg}")
        self.handeshake_socket.send(pickle.dumps(handshake_msg))
        self.log.debug(f"sent handshake message")


        if not self.handeshake_socket.poll(1000):
            raise Exception("Handshake socket not connected")

        handshake_response_buff = self.handeshake_socket.recv()
        handshake_response: FromLangServer = pickle.loads(handshake_response_buff)

        if (
            handshake_response["cmd"] != "connection_id"
            or handshake_response["data"] != self.connection["server_id"]
        ):
            raise Exception("Handshake response not correct")

    def initialize_channels(self):
        self.rep_server = RepChannel(self.socket, log=self.log)
        self.rep_server.trigger.connect(self.socket_reply)
        self.rep_server.start()
        self.log.debug("rep thread started")

        hb = HBChannel(self.handeshake_socket, log=self.log)
        self.hb_channels.append(hb)
        self.hb_channels[-1].trigger.connect(self.heartbeat_signal_handler)
        self.hb_channels[-1].start()
        self.log.debug("hb thread started")

    def restart_sockets(self):
        self.log.debug("Restarting sockets")
        self.stop_channels()
        self.close_sockets()
        self.sockets_created = False
        self.connection_id = None
        self.connected = False
        self.reconnect()

    def stop_channels(self):
        if self.rep_server:
            self.rep_server.stop()
        for hb in self.hb_channels:
            hb.stop()
        self.hb_channels = []

    def close_sockets(self):
        # if hasattr(self, 'socket'):
        self.socket.close()
        # if hasattr(self, 'handeshake_socket'):
        self.handeshake_socket.close()

    def handshake_send(self, msg: ToLangServer):
        if not self.sockets_created:
            self.connect_sockets()
            if not self.sockets_created:
                self.log.warning(f"handshake send failed, sockets not created {msg}")
                return

        buff = pickle.dumps(msg)
        self.handeshake_socket.send(buff)

    @Slot(bool)
    def set_connected(self, connected):
        self.connected = connected

    @Slot(str)
    def msg_routing(self, item):
        self.check_for_update()

    def check_for_update(self):
        data2 = get_db_data_new(self.db, self.app.file_search_path)
        self.log.debug(["checking for updated db data", data2])

        need_update = False

        for k, v in data2.items():
            if self.db_data[k] != v:
                need_update = True
                break

        if need_update:
            self.new_data = True
            self.db_data = data2
            self.send_db_data()
            self.app.update_tables()

    def start_lsp_logger(self):
        if self.app.debug:
            self.log.info("tell lang server debug mode")
            self.respond({"cmd": "debug", "data": True})

    def respond(self, msg: ToLangServer, no_block=False):
        if not self.sockets_created:
            self.connect_sockets()
            if not self.sockets_created:
                self.log.warning(f"Message send failed, sockets not created {msg}")
                return

        try:
            buff = pickle.dumps(msg)
            if self.connected:
                self.socket.send(buff, zmq.NOBLOCK if no_block else 0)
            else:
                self.log.error(f"Tried to send message when not connected {msg['cmd']}")
        except zmq.ZMQError as e:
            self.log.error(f"ZMQ error when sending message: {e}")
            self.restart_sockets()
        except Exception as e:
            self.log.error(f"Unexpected error when sending message: {e}")
            self.restart_sockets()

    def rec_manual(self):
        buff = self.socket.recv()
        data = pickle.loads(buff)
        self.log.debug(f"zmq server rec at start {data}")
        self.socket_reply(data)

    @Slot(object)
    def heartbeat_signal_handler(self, msg):
        if msg == "reset_sockets":
            self.restart_sockets()

    @Slot(object)
    def socket_reply(self, msg: FromLangServer):
        try:
            if not self.connected:
                self.connected = True
                self.start_lsp_logger()

            if msg["cmd"] == "connection_id":
                self.connection_id = msg["data"]

            elif msg["cmd"] == "run_sql":
                self.respond({"cmd": "run_sql", "msg": "success"})
                self.app.msg_routing(msg["sql"])

            elif msg["cmd"] == "db_data_update":
                self.send_db_data()
                self.start_lsp_logger()

            elif msg["cmd"] == "check_for_update":
                if self.new_data is True:
                    self.send_db_data()
                    self.new_data = False
                else:
                    self.respond({"cmd": "no_update"})

            elif msg["cmd"] == "heartbeat":
                self.hb = time.time()

            else:
                self.respond({"cmd": "not_handled"})

        except Exception as e:
            self.log.error(f"Error in socket_reply: {e}")
            self.restart_sockets()

    def send_db_data(self, no_block=False):
        reply = {"cmd": "db_data", "data": self.db_data}
        self.respond(reply, no_block=no_block)


from queue import Queue
import time
import logging
from dabbler.common import PprintSocketHandler
from logging.handlers import SocketHandler
from IPython import get_ipython
from qtpy import QtWidgets, QtCore, QtGui
import duckdb
from statistics import mean, quantiles
import sqlparse
from pathlib import Path
from pygments.formatters import HtmlFormatter
import datetime
from dabbler.gui_stuff import (
    check_dataframe_type,
    apply_fmt,
    format_map,
    fmt_types,
    gui_style,
    highlight,
)
from dabbler.gui_compenents import Shortcut, Vbox, Hbox, TreeItem, ZmqServer


def header_format(txt:str):
    if txt.startswith('ENUM'):
        return 'ENUM'
    return txt


class TableModel(QtCore.QAbstractTableModel):
    def __init__(
        self,
        data,
        selection_area: "TableSelectionArea",
        parent=None,
        hheaders: list[str] = None,
        htypes: list[str] = None,
    ):
        super(TableModel, self).__init__(parent)
        self._data = data
        self.selection_area = selection_area
        self.hheaders = hheaders
        self.htypes = htypes
        self.dispaly_role = QtCore.Qt.ItemDataRole.DisplayRole
        self.align_role = QtCore.Qt.ItemDataRole.TextAlignmentRole
        self.horiz = QtCore.Qt.Orientation.Horizontal

        self.alignment_map = {
            "STRING": QtCore.Qt.AlignmentFlag.AlignLeft
            + QtCore.Qt.AlignmentFlag.AlignVCenter,
            "NUMBER": QtCore.Qt.AlignmentFlag.AlignRight
            + QtCore.Qt.AlignmentFlag.AlignVCenter,
            "INT": QtCore.Qt.AlignmentFlag.AlignRight
            + QtCore.Qt.AlignmentFlag.AlignVCenter,
            "DATE": QtCore.Qt.AlignmentFlag.AlignRight
            + QtCore.Qt.AlignmentFlag.AlignVCenter,
            "DATETIME": QtCore.Qt.AlignmentFlag.AlignRight
            + QtCore.Qt.AlignmentFlag.AlignVCenter,
            "TIME": QtCore.Qt.AlignmentFlag.AlignRight
            + QtCore.Qt.AlignmentFlag.AlignVCenter,
        }

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._data[0]) if self.rowCount() else 0

    def data(self, index, role):
        if role == self.dispaly_role:
            row = index.row()
            if 0 <= row < self.rowCount():
                column = index.column()
                if 0 <= column < self.columnCount():
                    return self._data[row][column]
        if role == self.align_role:
            column = index.column()
            alignment = self.alignment_map.get(
                self.htypes[column],
                QtCore.Qt.AlignmentFlag.AlignLeft
                + QtCore.Qt.AlignmentFlag.AlignVCenter,
            )
            return alignment

    def headerData(self, section, orientation, role=QtCore.Qt.ItemDataRole.DisplayRole):
        if orientation == self.horiz and role == self.dispaly_role:
            return self.hheaders[section]
        return super().headerData(section, orientation, role)


class DisplayArea(QtWidgets.QWidget):
    def __init__(self, parent=None, parent_layout=None):
        super().__init__(parent)
        self.lay_out = Vbox(self)
        self.lay_out.addWidget(QtWidgets.QLabel("Display Area"))
        self.table = QtWidgets.QTableView()

        self.lay_out.addWidget(self.table)
        # self.table.setMinimumWidth(800)
        # self.table.setMinimumHeight(450)
        self.table.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents
        )
        self.table.verticalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents
        )
        self.table.verticalHeader().setVisible(False)


class TableSelectionArea(QtWidgets.QWidget):
    def __init__(
        self,
        parent: "mainWindow" = None,
        db: duckdb.DuckDBPyConnection = None,
        parent_layout=None,
        table: QtWidgets.QTableView = None,
        col_selector: "ColSelectionArea" = None,
    ):
        super().__init__(parent)
        self.main = parent
        self.db = db
        self.table = table
        self.table_list: dict[str, str] = {}
        # self.setMaximumWidth(275)
        self.lay_out = Vbox(self)
        self.db_size = QtWidgets.QLabel("DB size: ")
        self.lay_out.addWidget(self.db_size)
        self.query: str = None
        self.selected_table: str = None
        self.column_types: dict[str, dict[str, str]] = {}
        self.table_selection = QtWidgets.QTreeWidget()
        self.table_selection.setColumnCount(2)
        self.table_selection.setHeaderLabels(["Name", "Type"])
        self.table_selection.setColumnWidth(0, 200)
        self.table_selection.setColumnWidth(1, 50)

        self.table_selection.itemSelectionChanged.connect(self.select_table)
        self.df_list = set()

        self.model = self.table.model()
        self.lay_out.addWidget(self.table_selection)

        self.populate_table_list()

    def populate_table_list(self):
        self.table_selection.clear()
        mem, mem_limit = self.db.execute(
            "select memory_usage, memory_limit from pragma_database_size()"
        ).fetchone()
        self.db_size.setText(f"DB Size: {mem}/{mem_limit}")
        self.current_schema = self.db.execute("select current_schema()").fetchone()[0]
        self.current_db = self.db.execute("select current_database()").fetchone()[0]

        tbl_data = self.db.execute(
            """--sql
            with cte as (
                select "database_name", schema_name, table_name, 'table' as table_type from duckdb_tables
                union all
                select "database_name", schema_name, view_name, 'view' as table_type from duckdb_views
                union all
                select 'system_info', schema_name, function_name || '()', 'table function' 
                    from duckdb_functions() 
                    where function_type = 'table' and parameters = []
                    and function_name not in ('checkpoint','test_all_types','force_checkpoint','index_scan','seq_scan')
                union all
                select distinct 'system_info', "database_name" || '.' || schema_name, table_name, 'table' 
                    from duckdb_columns() 
                    where schema_name in ('information_schema', 'pg_catalog')
                    and database_name not in ('system','temp')            )
            select * from cte order by "database_name", schema_name, table_name
            """
        ).fetchall()

        dbs: dict[str, TreeItem] = {}
        dbs[self.current_db] = TreeItem([f"{self.current_db}"], item_type="db")

        for db, scm, tbl, typ in tbl_data:
            if db not in dbs:
                dbs[db] = TreeItem([f"{db}"], item_type="db")
            dbs[db].addChild(TreeItem([f"{scm}.{tbl}", typ], item_type=db))

        if not self.main.app.in_thread:
            dbs["dataframes"] = TreeItem(["dataframes"], item_type="db")
            ipython = get_ipython()
            self.df_list = set()
            for item in ipython.ev("dir()"):
                i_type = str(type(ipython.ev(item)))
                df_type = check_dataframe_type(i_type)
                
                if df_type and item[0] != "_":
                    dbs["dataframes"].addChild(
                        TreeItem(
                            [item, df_type],
                            item_type=df_type,
                        )
                    )
                    self.df_list.add(item)
                    self.table_list[item] = df_type

        added_items = [self.current_db, "dataframes"]
        sys_info = dbs.pop("system_info")

        items = []
        for x in added_items:
            if x in dbs:
                items += [dbs[x]]

        items += [v for k, v in dbs.items() if k not in added_items]
        items += [sys_info]

        self.table_selection.addTopLevelItems(items)
        self.table_selection.setIndentation(7)

        for k, db in [(k, db) for k, db in dbs.items() if k != "system_info"]:
            db.setExpanded(True)

    def select_table(self):
        item = self.table_selection.selectedItems()[0]
        if item.item_type == "db":
            return

        self.selected_table = item.data(0, 0)
        tbl_db = f"{item.parent().data(0,0)}."
        if tbl_db == "system_info." or tbl_db == "dataframes.":
            tbl_db = ""

        self.query_db(f"""select * from {tbl_db}{self.selected_table} """, True)

    def query_db(_self_z, _stmt_z: str, _tbl_click=False):
        _self_z.db = _self_z.db.cursor()
        _self_z.db.execute(f"set file_search_path to '{_self_z.main.app.file_search_path}'")

        # if _self_z.selected_table in _self_z.df_list and _tbl_click:
        #     _ip_z = get_ipython()
        #     locals().__setitem__(_self_z.selected_table,_ip_z.ev(_self_z.selected_table))
        #     if _self_z.table_list[_self_z.selected_table] == 'duckdb_rel':
        #         for _t_z_ in _self_z.table_list:
        #             locals().__setitem__(_t_z_,_ip_z.ev(_t_z_))

        _ipython_z = get_ipython()
        for _item_z in _ipython_z.ev("dir()"):
            _i_type_z = str(type(_ipython_z.ev(_item_z)))
            _df_type = check_dataframe_type(_i_type_z)
            
            
            if _df_type and _item_z[0] != "_":
                locals().__setitem__(_item_z, _ipython_z.ev(_item_z))

        try:
            _start = time.time()
            
            _desc = _self_z.db.execute(f"Describe {_stmt_z}").fetchall()
            
            # desc_stmt = f"{_self_z.main.app.db_name}.execute('''Describe {_stmt_z}''').fetchall()"
            # _self_z.main.app.log.info(desc_stmt)
            # _desc = _ipython_z.ev(desc_stmt)
            
            _self_z.headers = [f"{x[0]}\n{header_format(x[1])}" for x in _desc]
            _self_z.dtypes = [x[1] for x in _desc]
            _self_z.htypes = [fmt_types.get(x, "STRING") for x in _self_z.dtypes]
            
            _self_z.result = _self_z.db.execute(_stmt_z)
            
            # exec_stmt = f"{_self_z.main.app.db_name}.execute('''{_stmt_z}''')"
            # _self_z.result = _ipython_z.ev(exec_stmt)
            
            
            
            _self_z.query = _stmt_z
            _duration = time.time() - _start
            _self_z.main.status.setText(f"{_duration:.2f} sec to execute")

            # highlighted_query_text = highlight(sqlparse.format(_stmt_z,reindent_aligned=True))
            _highlighted_query_text = highlight(_stmt_z)
            _self_z.main.sql_doc.setHtml(_highlighted_query_text)
            _self_z.next_page()

        except Exception as e:
            _self_z.main.status.setText(f"Exception executing query:\n{e}")
            _highlighted_query_text = highlight(
                sqlparse.format(_stmt_z, reindent_aligned=True)
            )
            _self_z.main.sql_doc.setHtml(_highlighted_query_text)


    def get_col_fmt(self, col_type, col_index, data):
        try:
            fmt_type = fmt_types.get(col_type, "STRING")
            if fmt_type == "NUMBER":
                col_abs = [
                    abs(x[col_index])
                    for x in data
                    if x[col_index] is not None and x[col_index] != 0
                ]
                if len(col_abs) == 0:
                    return ""

                if len(col_abs) == 1:
                    col_avg = col_abs[0]
                else:
                    col_avg = mean(quantiles(col_abs, n=10)[1:-1])

                if col_avg < 1:
                    return ".5f"
                if col_avg < 5:
                    return ".2f"
                if col_avg < 99:
                    return ".1f"
                return ",.0f"
            if fmt_type == "INT":
                col_abs = [
                    abs(x[col_index])
                    for x in data
                    if x[col_index] is not None and x[col_index] != 0
                ]
                if len(col_abs) == 0:
                    return ""

                if len(col_abs) == 1:
                    col_avg = col_abs[0]
                else:
                    col_avg = mean(quantiles(col_abs, n=10)[1:-1])

                if col_avg < 3000:
                    return ""
                if col_avg > 3000:
                    return ","

            fmt_str = format_map.get(fmt_type, "")
        except Exception as e:
            self.main.status.setText(
                f"fmt exception:{e}\ncol_type:{col_type}\ndata:{[x[col_index] for x in data]}"
            )
        return fmt_str

    def next_page(self):
        try:
            subtract_amt = 0
            if self.table.horizontalScrollBar().isVisible():
                subtract_amt = 24
            rows_to_fetch = int((self.height() - 48 - subtract_amt) / 23)
            self.model = self.table.model()

            if self.model is not None:
                self.table.setModel(None)
                self.model.deleteLater()

            data = self.result.fetchmany(rows_to_fetch)

            if len(data) == 0:
                self.main.status.setText("end of query")
                return

            fmts = [
                self.get_col_fmt(col_type, i, data)
                for i, col_type in enumerate(self.dtypes)
            ]

            # data = [[f' {x:{fmts[c]}} ' for c,x in enumerate(row)] for row in data]
            data = [[apply_fmt(x, c, fmts) for c, x in enumerate(row)] for row in data]

            self.model = TableModel(data, self.table, self, self.headers, self.htypes)
            self.table.setModel(self.model)
            self.table.setCurrentIndex(self.table.model().index(0, 0))

        except Exception as e:
            self.main.status.setText(f"Exeption on next_page:\n{e}\nfmts:{fmts}")


class ColSelectionArea(QtWidgets.QWidget):
    def __init__(
        self,
        parent=None,
        db: duckdb.DuckDBPyConnection = None,
        parent_layout=None,
        table: QtWidgets.QTableView = None,
    ):
        super().__init__(parent)
        self.db = db
        self.table = table
        self.setMaximumWidth(150)
        self.lay_out = Vbox(self)
        self.lay_out.addWidget(QtWidgets.QLabel("Tbl Columns"))
        self.selected_table: str = None
        self.column_types: dict[str, list[(str, str)]] = {}
        self.col_sel_model = QtGui.QStandardItemModel()
        self.col_selection = QtWidgets.QListView(self)
        self.col_selection.setModel(self.col_sel_model)
        self.col_selection.setEditTriggers(
            QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers
        )
        self.col_selection.selectionModel().selectionChanged.connect(
            self.populate_table
        )
        self.lay_out.addWidget(self.col_selection)


class TopBar(QtWidgets.QWidget):
    def __init__(self, parent=None, parent_layout=None):
        super().__init__(parent)
        self.lay_out = Hbox(self)


class mainWindow(QtWidgets.QWidget):
    def __init__(
        self,
        db: duckdb.DuckDBPyConnection,
        app: "MyApp",
        parent=None,
    ):
        super().__init__(parent)
        self.db = db
        self.app = app
        self.split = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical)
        self.setWindowTitle("dabbler")
        self.top_layout = Vbox(self)
        self.top_bar = TopBar(self, self.top_layout)
        self.top_layout.addWidget(self.top_bar)
        self.top_layout.addWidget(self.split)

        self.mid_layout = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        self.split.addWidget(self.mid_layout)
        self.display_area = DisplayArea(self, self.mid_layout)
        self.selection_area = TableSelectionArea(
            self, self.db, self.mid_layout, self.display_area.table
        )
        self.mid_layout.addWidget(self.selection_area)
        self.mid_layout.addWidget(self.display_area)
        self.status = QtWidgets.QLabel()
        self.sql_doc = QtGui.QTextDocument()
        self.sql_doc.setDefaultStyleSheet(HtmlFormatter().get_style_defs())
        self.sql_textbox = QtWidgets.QTextEdit()
        self.sql_textbox.setReadOnly(True)
        self.sql_textbox.setDocument(self.sql_doc)
        self.status.setText("Ready")
        self.bottom_section = QtWidgets.QTabWidget()
        self.bottom_section.setMinimumHeight(100)
        self.bottom_layout = Vbox(self.bottom_section)
        self.bottom_layout.addWidget(self.status)
        self.bottom_layout.addWidget(self.sql_textbox)
        self.split.addWidget(self.bottom_section)

        btn_new_window = QtWidgets.QPushButton()
        btn_new_window.setText("new window")
        btn_new_window.clicked.connect(self.app.new_window)
        self.top_bar.lay_out.addWidget(btn_new_window)

        btn_to_excel = QtWidgets.QPushButton()
        btn_to_excel.setText("open_in_excel")
        btn_to_excel.clicked.connect(self.to_excel)
        self.top_bar.lay_out.addWidget(btn_to_excel)

        btn_refresh_tables = QtWidgets.QPushButton()
        btn_refresh_tables.setText("update table list")
        btn_refresh_tables.clicked.connect(self.selection_area.populate_table_list)
        self.top_bar.lay_out.addWidget(btn_refresh_tables)

        self.short_keys = [
            # Shortcut('Alt+Right',self,lambda:checked_radio[0].next.setChecked(True)),
            # Shortcut('Escape',self,self.hide_window),
            Shortcut("PgDown", self, self.selection_area.next_page),
        ]
        width = 960
        self.setGeometry(900, 200, width, 600)
        self.activateWindow()
        tabl_sel_width = 285
        self.mid_layout.setSizes([tabl_sel_width, width - tabl_sel_width])
        self.mid_layout.setStretchFactor(1, 1)

    def to_excel(self):
        if not self.selection_area.query:
            self.status.setText("query not found")
            return
        try:
            self.status.setText("clicked to excel")
            Path(r"c:\temp_rpt").mkdir(exist_ok=True)
            self.status.setText("make_path")

            output_file = Path(r"c:\temp_rpt").joinpath(
                f"query_{datetime.datetime.now():%Y%m%d%M%S}.csv"
            )
            output = self.db.sql(self.selection_area.query)
            self.status.setText("db_run")
            output.limit(100000).to_csv(str(output_file), header=True)
            self.status.setText("csv_saved")
            import win32com.client as win

            excel = win.Dispatch("Excel.Application")
            excel.Workbooks.Open(str(output_file))
            excel.Visible = True
            self.status.setText("excel_open")
        except Exception as e:
            self.status.setText(f"Exception trying to export to excel:\n{e}")


class MyApp(QtWidgets.QApplication):
    def __init__(
        self, argv, db=None, file_search_path = None, debug = False, file:str = None, db_name:str = None
        
    ) -> None:
        super().__init__(argv)
        self.db = db
        self.db_name = db_name
        self.windows: list[mainWindow] = []
        self.in_thread = False
        self.file_search_path = file_search_path
        self.py_file:Path = Path(file)
        self.stylesheet = gui_style
        self.debug = debug
        self.log = logging.getLogger("dabbler_gui")
        if self.debug:
            self.log.setLevel(1)
            socket_handler = PprintSocketHandler('127.0.0.1', 19996)
            # socket_handler.setFormatter(LogFmt)
            self.log.addHandler(socket_handler)
            self.log.info("debugging")
        self.zmq = ZmqServer(self, self.db)
        # self.q_handler = MsgHandler(q=q,signal_to_emit=self.msg_routing)
        # self.focusWindowChanged.connect(self.update_tables)
        self.setStyleSheet(gui_style)
        self.new_window()
        # if in_thread:
        # self.exec()

    # def update_tables(self):
    #     self.windows = [w for w in self.windows if w.isVisible()]

    #     for w in self.windows:
    #         w.selection_area.populate_table_list()

    # @Slot(str)
    def msg_routing(self, item):
        self.windows[0].selection_area.query_db(item)

    def update_tables(self):
        self.windows = [w for w in self.windows if w.isVisible()]
        for w in self.windows:
            w.selection_area.populate_table_list()

    def new_window(self):
        window = mainWindow(self.db, self)
        window.show()
        # window.move(QtCore.QPoint(900,200))
        self.windows.append(window)
        for w in self.windows:
            if not w.isVisible():
                w.deleteLater()
        self.windows = [w for w in self.windows if w.isVisible()]

        txt = f"windows open: {len(self.windows)}"
        # for w in self.windows:
        # txt += f'\n{w},{w.isVisible()}'
        window.status.setText(txt)



# from dabbler.lsp import parser
# from dabbler.lsp import server_classes
# from dabbler.lsp import sql_utils
# from dabbler.lsp import db_data
# from dabbler.lsp import completion

# from . import parser
# from . import server_classes
# from . import sql_utils
# from . import db_data
# from . import completion


import os
import sys
import logging
from pathlib import Path
from logging.handlers import SocketHandler
from functools import partial
import re as regex
import zmq
from dataclasses import dataclass
from zmq.asyncio import Context, Socket as AsyncSocket, Poller
from pygls.server import LanguageServer
import pickle
import asyncio
from dabbler.lsp.db_data import make_db, make_completion_map
from dabbler.lsp.sql_utils import strip_sql_whitespace, SelectNode, CmpItem
from dabbler.lsp.completion import (
    PathCompleter,
    duckdb_extensions,
    duckdb_settings,
    duckdb_pragmas,
    duckdb_types,
    file_path_completion_regex,
)
from dabbler.lsp.parser import SqlParserNew
from dabbler.common import FromLangServer, ToLangServer, KeyFile, PprintSocketHandler, grammer_kw

# from dabbler.lsp.completer import CmpItem
from lsprotocol.types import (
    CompletionItem,
    CompletionList,
    CompletionItemKind,
)


# %%


class InlineSqlLangServer(LanguageServer):
    CMD_SEND_SQL_TO_GUI = "sendSqlToDbDabbler"
    CMD_FORMAT_CURRENT_STATEMENT = "dbDabblerFormatCurrentStatement"
    CONFIGURATION_SECTION = "pygls.jsonServer"

    def __init__(self, *args):
        super().__init__(*args)

        self.log = logging.getLogger("dabbler_lsp")
        self.debug = False
        self.ctx = Context()
        self.create_sockets2()

        self.completer: "SqlCompleter" = None
        self.socket_connected = False
        self.socket_created = False
        self.pathlibs_paths = {}
        self.key_file: KeyFile = None

    def create_default_compelter(self):
        if not self.completer is None:
            return
        blank_db_data = {
            'data':[],
            'dataframes':[],
            'databases':[],
            'functions':[],
            'paths':[],
            'schemas':[],
            'current_schema':'memory.main',
            'cwd':self.workspace.root_path,
            'file_search_path':None,
        }
        self.completer = SqlCompleter(blank_db_data, self)

    def start_io(self, stdin = None, stdout = None):

        self.loop.create_task(self.zmq_recv(self.poller))
        # self.loop.run_until_complete(self.zmq_recv(self.poller))
        super().start_io(stdin, stdout)
        
    def start_logging(self):
        if self.debug:
            return
        self.debug = True
        self.log.setLevel(1)  # to send all records to cutelog
        socket_handler = PprintSocketHandler(
            "127.0.0.1", 19996
        )  # default listening address
        self.log.addHandler(socket_handler)
        self.log.info("logging started")
        self.log.debug(os.environ)

    def log_workspace_info(self):
        if self.debug is False:
            return
        if self.workspace is None:
            self.log.debug(f"workspace is None")
            return
        workspace_info = {
            "root_uri": self.workspace.root_uri,
            "root_path": self.workspace.root_path,
            "folders": self.workspace.folders,
            "documents": self.workspace.documents,
        }
        self.log.debug(["workspace_info", workspace_info])
        
    def find_port(self):
        # ctx = zmq.Context()
        socket = self.ctx.socket(zmq.PAIR)
        port = socket.bind_to_random_port("tcp://127.0.0.1")
        socket.close()
        return port
    


    def create_sockets2(self):
        
        # ctx1 = Context().instance()
        self.socket = self.ctx.socket(zmq.PAIR)
        self.main_port = self.find_port()
        self.socket.connect(f"tcp://127.0.0.1:{self.main_port}")

        # ctx2 = Context().instance()
        self.handshake_socket = self.ctx.socket(zmq.PAIR)
        self.handshake_port = self.find_port()
        self.handshake_socket.bind(f"tcp://127.0.0.1:{self.handshake_port}")

        self.poller = Poller()
        self.poller.register(self.socket, zmq.POLLIN)
        self.poller.register(self.handshake_socket, zmq.POLLIN)
        # self.loop.create_task(self.zmq_recv(self.poller))
        
        self.log.debug(f"main_port {self.main_port}, handshake_port {self.handshake_port}")

    def save_key_file(self):
        if self.key_file:
            return
        if self.workspace is None:
            self.log.debug(f"workspace is None, keyfile not saved")
            return
        self.key_file = KeyFile()
        
        workpace_path = str(self.workspace.root_path)
        
        connection = self.key_file.add_connection(
            workpace_path,
            {
                "workspace_path": workpace_path,
                "main_port": self.main_port,
                "handshake_port": self.handshake_port,
            },
        )
        self.connection_info = connection
        self.show_message_log(f"main_port {self.main_port}, handshake_port {self.handshake_port}")
        self.log.debug(f"key_file {self.key_file.connections}")

        self.zmq_send({"cmd": "db_data_update","con_id":self.connection_info['server_id']})

    def initialize_sql_completer(self, data: dict):
        self.completer = SqlCompleter(data, self)
        self.pathlibs_paths = {x[0]:x[1] for x in data["paths"]}
        self.show_message_log("SqlCompleter initialized")

    async def handle_zmq_message(self, msg: ToLangServer):
        if msg["cmd"] == "db_data":
            self.show_message_log("received db_data")
            await self.loop.run_in_executor(
                self.thread_pool_executor,
                partial(self.initialize_sql_completer, msg["data"])
            )
        elif msg["cmd"] == "ip_python_started":
            self.show_message_log(f"ip_python_started = {msg['data']}")
            self.zmq_send({"cmd": "db_data_update"})
        elif msg["cmd"] == "ip":
            self.show_message_log(f"ipython event = {msg['data']}")
        elif msg["cmd"] == "no_update":
            self.show_message_log("check update: no update")
        elif msg["cmd"] == "debug":
            if msg["data"]:
                self.show_message_log("started in debug mode")
                self.start_logging()
        elif msg["cmd"] == "heartbeat":
            await self.handshake_socket.send(pickle.dumps({"cmd": "heartbeat"}))
        elif msg["cmd"] == "connection_id":
            con_id = msg["data"]
            if con_id == self.connection_info["client_id"]:
                await self.handshake_socket.send(pickle.dumps({"cmd": "connection_id", "data": self.connection_info["server_id"]}))
            else:
                self.log.debug(f"connection_id {con_id} != {self.connection_info['client_id']}")
                self.show_message_log(f"connection_id {con_id} != {self.connection_info['client_id']}")


    async def zmq_recv(self, poller: Poller):
        while self._stop_event is None:
            await asyncio.sleep(0.1)
        stop = self._stop_event

        while not stop.is_set():
            try:
                socks: dict[AsyncSocket, AsyncSocket] = dict(await poller.poll())
                for socket in socks:
                    buff = await socket.recv()

                    if not self.socket_connected:
                        self.socket_connected = True
                        self.show_message_log(f"connected to IPython")
                        self.log.info("connected to IPython")
                    
                    msg: ToLangServer = pickle.loads(buff)
                    await self.handle_zmq_message(msg)
            
            except Exception as e:
                self.log.error(f"Error in zmq_recv: {str(e)}", exc_info=True)
                await asyncio.sleep(1)  # Prevent tight error loop


    def zmq_send(self, msg:FromLangServer, no_block=False):
        if no_block:
            try:
                msg = pickle.dumps(msg)
                self.socket.send(msg, zmq.NOBLOCK)
            except:  # noqa: E722
                self.show_message_log("zmq send failed")
            return
        if self.socket_connected:
            # self.show_message_log(f"zmq send {msg}")
            msg = pickle.dumps(msg)
            self.socket.send(msg)

    def zmq_check_for_update(self):
        if self.socket_connected:
            self.zmq_send({"cmd": "check_for_update"})
        else:
            self.show_message_log("zmq not connected")


table_types = set(["table", "database", "schema", "cte"])


@dataclass
class PasredItemsCache:
    age: int
    last_line: int
    rng_start: int
    spaces_on_cur_line: int
    items: dict[str, list[CmpItem]]


class SqlCompleter:
    def __init__(self, db_data, ls: InlineSqlLangServer = None) -> None:
        self.db = make_db(db_data)
        self.completion_map = make_completion_map(self.db, db_data)
        self.db_data = db_data
        if Path(db_data["cwd"]).is_dir():
            os.chdir(db_data["cwd"])
        self.path_completer = PathCompleter(db_data["cwd"], db_data["file_search_path"],ls.log.getChild('path_completer'))
        self.ls = ls
        self.file_search_path = db_data["file_search_path"]
        self.log = ls.log.getChild("completer")
        self.log_comp_map = self.log.getChild("comp_map")
        # self.parsed_times_cache = PasredItemsCache(99,0,0,0,{})
        self.parser2 = SqlParserNew(self.db, self.ls, ls.log,self.file_search_path)

    def show_message_log(self, msg):
        if self.ls:
            self.ls.show_message_log(msg)

    def get_queries(self, pos, sql):
        queries, choices_pos = self.parser2.parse_sql(sql, pos)
        if not queries:
            return None, None, choices_pos
        queries.queries_list.sort(key=lambda x: x.end_pos - x.start_pos)
        if len(queries.queries_list) == 0:
            return None, None, choices_pos
        filtered = [x for x in queries.queries_list if x.start_pos <= pos <= x.end_pos]
        if len(filtered) == 0:
            return queries.queries_list[-1], queries, choices_pos
        q = filtered[0]
        return q, queries, choices_pos

    def parse_sql2(self, pos, sql: str):
        comp_map = {}
        comp_map["root_namespace"] = []

        q, queries, choices_pos = self.get_queries(pos, sql)
        
        if choices_pos:
            kw_comps = []
            for c in choices_pos:
                if c in grammer_kw:
                    kw_comps.append(
                        CmpItem(c, CompletionItemKind.Keyword, None, "keyword", "3", "keyword")
                    )
            
            comp_map["root_namespace"].extend(kw_comps)
        
        
        if q is None:
            # self.log_comp_map.debug(comp_map)
            return comp_map
        for k, v in q.from_refs.items():
            if v.kind.name == "subquery":
                projection = queries.queries[v.start_pos].projection
                comp_map[k] = [
                    CmpItem(x[0], CompletionItemKind.Field, None, x[1], "1", "column")
                    for x in projection
                ]
                comp_map["root_namespace"].append(
                    CmpItem(k, CompletionItemKind.File, None, "sub query", "1", "table")
                )
                continue
            
            if v.kind.name == "table_function":
                projection = v.projection
                if not projection:
                    continue
                comp_map[k] = [
                    CmpItem(x[0], CompletionItemKind.Field, None, x[1], "1", "column")
                    for x in projection
                ]
                comp_map["root_namespace"].append(
                    CmpItem(k, CompletionItemKind.File, None, "table function", "1", "table")
                )
                continue
            
            # self.show_message_log(f'{v}')
            if q.ctes and v.name in q.ctes.map:
                comp_map[k] = [
                    CmpItem(x[0], CompletionItemKind.Field, None, x[1], "1", "column")
                    for x in q.ctes.map[v.name].projection
                ]
                comp_map["root_namespace"].append(
                    CmpItem(k, CompletionItemKind.File, None, "cte", "1", "cte")
                )
                continue
            elif v.name in q.cte_sibblings:
                comp_map[k] = [
                    CmpItem(x[0], CompletionItemKind.Field, None, x[1], "1", "column")
                    for x in q.cte_sibblings[v.name].projection
                ]
                comp_map["root_namespace"].append(
                    CmpItem(k, CompletionItemKind.File, None, "cte", "1", "cte")
                )
                continue
            else:
                if v.name not in self.completion_map:
                    continue
                comp_map[k] = self.completion_map[v.name]
                comp_map["root_namespace"].append(
                    CmpItem(
                        k, CompletionItemKind.File, None, "table_alias", "1", "table"
                    )
                )

        if q.ctes:
            for k, v in q.ctes.map.items():
                comp_map[k] = [
                    CmpItem(x[0], CompletionItemKind.Field, None, x[1], "1", "column")
                    for x in v.projection
                ]
                comp_map["root_namespace"].append(
                    CmpItem(k, CompletionItemKind.File, None, "cte", "1", "cte")
                )

        for k, v in q.cte_sibblings.items():
            comp_map[k] = [
                CmpItem(x[0], CompletionItemKind.Field, None, x[1], "1", "column")
                for x in v.projection
            ]
            comp_map["root_namespace"].append(
                CmpItem(k, CompletionItemKind.File, None, "cte", "1", "cte")
            )

        # self.show_message_log(f'cte_sibblings {q.cte_sibblings}')

        col_to_add = []
        labels_added = []

        for k in comp_map["root_namespace"]:
            if k in q.from_refs and k in comp_map:
                col_to_add.extend(
                    [x for x in comp_map[k] if x.label not in labels_added]
                )
                labels_added.extend([x.label for x in comp_map[k]])

        comp_map["root_namespace"].extend(col_to_add)

        # self.log_comp_map.debug(comp_map)
        

        
        return comp_map

    # def get_comp_map(self, cursor_pos, sql_rng: SelectNode):
    #     parsed_items = self.parse_sql2(cursor_pos, sql_rng.txt)
    #     comp_map: dict[str, list[CmpItem]] = {}
    #     comp_map.update(parsed_items)
    #     comp_map.update(self.completion_map)
    #     return comp_map

    def route_completion2(
        self,
        cursor_pos: int,
        sql_rng: SelectNode,
        trigger: str,
        current_line: int,
        current_line_txt: str,
    ):
        # self.show_message_log(f'route_completion pos:{cursor_pos}, trigger:{trigger}, line:{current_line}, line_txt{current_line_txt}')
        # self.parsed_times_cache.age += 1
        sql_left_of_cur = strip_sql_whitespace(sql_rng.txt[:cursor_pos])
        # self.show_message_log(f'comp_map_size: {sys.getsizeof(self.completion_map)} parser: {sys.getsizeof(self.parser)}, trigger: {trigger}')

        if sql_left_of_cur[-2:] == "::":
            return CompletionList(is_incomplete=False, items=duckdb_types)

        if regex.match(
            "(^| )(load |install )$", sql_left_of_cur, flags=regex.IGNORECASE
        ):
            return CompletionList(is_incomplete=False, items=duckdb_extensions)

        if regex.match("(^| )(pragma )$", sql_left_of_cur, flags=regex.IGNORECASE):
            return CompletionList(is_incomplete=False, items=duckdb_pragmas)

        if regex.match("(^| )(set |reset )$", sql_left_of_cur, flags=regex.IGNORECASE):
            return CompletionList(
                is_incomplete=False,
                items=duckdb_settings
                + [CompletionItem(label=c) for c in ["LOCAL", "SESSION", "GLOBAL"]],
            )

        if regex.match(
            "(^| )(set |reset )(local | session | global )$",
            sql_left_of_cur,
            flags=regex.IGNORECASE,
        ):
            return CompletionList(is_incomplete=False, items=duckdb_settings)

        file_comp = file_path_completion_regex.match(sql_left_of_cur)
        if file_comp:
            items = self.path_completer.get_items(file_comp.group(2))
            return CompletionList(is_incomplete=False, items=items)


        try:
            parsed_items: dict[str, list[CmpItem]] = self.parse_sql2(
                cursor_pos, sql_rng.txt
            )
        except Exception as e:
            # self.log.info(["parsed_items_error", e,sys.exception().__traceback__])
            # self.log.debug('parsed_items_error')

            parsed_items = {"root_namespace": []}
        # self.show_message_log(f'{parsed_items}')
        # comp_map:dict[str,list[CmpItem]] = {}
        # comp_map.update(parsed_items)
        # comp_map.update(self.completion_map.copy())
        # comp_map['root_namespace'].extend(parsed_items['root_namespace'])
        comp_map = self.completion_map

        if regex.match(
            ".*(^| )(join |from |pivot |unpivot |alter table |insert into )(\w+( \w+)?, )*\w?$",
            sql_left_of_cur,
            flags=regex.IGNORECASE,
        ):
            tbls = [
                x.comp for x in comp_map["root_namespace"] if x.obj_type in table_types
            ]
            tbls += [
                x.comp
                for x in parsed_items["root_namespace"]
                if x.obj_type in table_types
            ]
            # ls.show_message_log(f'join {tbls}')
            return CompletionList(is_incomplete=False, items=tbls)

        if trigger == " ":
            return

        m = regex.match(
            ".*(^| )(join |from |pivot |unpivot |alter table |insert into )(\w+( \w+)?, )*(?P<dotitems>(\w+\.)+)$",
            sql_left_of_cur,
            flags=regex.IGNORECASE,
        )
        if m and trigger == ".":
            key = m.group("dotitems").strip(".")
            items = []
            if key in comp_map:
                items = [x.comp for x in comp_map[key] if x.obj_type in table_types]
            if key in parsed_items:
                items += [
                    x.comp for x in parsed_items[key] if x.obj_type in table_types
                ]
            return CompletionList(is_incomplete=False, items=items)

        m = regex.match(
            ".*(^| |\()(?P<dotitems>(\w+\.)+)$", sql_left_of_cur, flags=regex.IGNORECASE
        )
        if m and trigger == ".":
            key = m.group("dotitems").strip(".")
            items = []
            if key in comp_map:
                items += [x.comp for x in comp_map[key]]
            if key in parsed_items:
                items += [x.comp for x in parsed_items[key]]
            if len(items) > 0:
                return CompletionList(is_incomplete=False, items=items)

        m = regex.match(
            ".*(^| |\()(?P<char>(\w+))$", sql_left_of_cur, flags=regex.IGNORECASE
        )
        if m:
            # comp_items = []
            # comp_items.extend(self.completion_map['root_namespace'])
            # comp_items.extend(parsed_items['root_namespace'])
            chars = m.group("char").lower()
            comp_items = [x for x in comp_map["root_namespace"]]
            comp_items += parsed_items["root_namespace"]
            return CompletionList(
                is_incomplete=True,
                items=[x.comp for x in comp_items if x.label.lower().startswith(chars)],
            )

        return None



#%%
from dataclasses import dataclass
from itertools import chain
import duckdb
import sqlparse
import re as regex
from dabbler.common import check_name

from lsprotocol.types import (
    CompletionItem,
    CompletionItemKind,
    CompletionItemLabelDetails,
)

@dataclass
class SelectNode:
    stmt:sqlparse.sql.Parenthesis
    txt:str
    length:int
    start:int
    end:int
    cur_idx:int
    

@dataclass
class SqlStatement:
    stmt:sqlparse.sql.Statement
    length:int
    start:int
    end:int
    txt:str
    cur_idx:int
    

@dataclass
class SqlTxtRange:
    txt:str
    length:str
    start:str
    end:str
    cur_idx:int


@dataclass
class CmpItem:
    label:str
    kind:int
    detail:str
    typ:str
    sort:str
    obj_type:str
    doc:str = None
    
    def __eq__(self, other) -> bool:
        return self.label == other
    
    def __hash__(self) -> int:
        return hash(self.label)
    
    def __repr__(self) -> str:
        return self.label
    
    @property    
    def comp(self):
        if self.kind == CompletionItemKind.Keyword:
            label = self.label
        else:
            label = label=check_name(self.label.strip('"'))

        if self.typ and self.typ.startswith('ENUM('):
            description = 'ENUM'
        else:
            description = self.typ
        
        return CompletionItem(
            label=label,
            kind=self.kind,
            sort_text=self.sort,
            filter_text=self.label.lower(),
            documentation=self.doc,
            label_details=CompletionItemLabelDetails(
                description=description),
                detail=self.detail,
            )


def line_col(str, idx):
    return str.count('\n', 0, idx) + 1, idx - str.rfind('\n', 0, idx)


def strip_sql_whitespace(txt):
    txt = regex.sub('(.*)--sql.*\n','\g<1>',txt,flags=regex.IGNORECASE)
    txt = regex.sub('\s+',' ',txt,flags=regex.IGNORECASE)
    return txt


select_node_words = [
    'select',
    'from',
    'pivot',
    'unpivot',
]

whitespace_tokens = [
    sqlparse.tokens.Newline,
    sqlparse.tokens.Whitespace,
]


def get_stmts(sql_txt,pos_in_doc):
    stmts:list[SqlStatement] = []
    
    for x in sqlparse.parse(sql_txt):
        
        start=sql_txt.find(x.value)+pos_in_doc
        length = len(x.value)
        
        stmt = SqlStatement(
            stmt=x,
            length=length,
            start=start,
            end=start+length,
            txt=x.value,
            cur_idx=pos_in_doc
        )
        
        stmts.append(stmt)
    return stmts


def get_statement(rng:SqlTxtRange,txt:str) -> SqlStatement:
    
    pos = rng.cur_idx
    result = None
    length = len(rng.txt)
    
    for stmt in sqlparse.parse(rng.txt):
        
        skipped = 0
        pattern = "(create\s+)(or\s+replace\s+)?(?P<type>view|table)\s+(?P<name>\w+)(\s+as\s+)(?P<select>.*)"
        match = regex.search(pattern,stmt.value,regex.DOTALL | regex.IGNORECASE)
        if match:
            # print(match.group('select'))
            stmt = sqlparse.parse(match.group('select'))[0]
            skipped = match.span('select')[0]
            # print(stmt.value)
        
        s = rng.txt.find(stmt.value)+rng.start
        e = s + len(stmt.value)
        l = e - s
        # print(txt[s:e])
        # print(f'pos:{pos} se:{s},{e}')
                
        # print(pos,s,e)
        if pos >= s and pos <= e and l <= length:
            result = stmt
            length = l
            r_s, r_e, r_l = s,e,l
            
            
    if result:
        return SqlStatement(
            stmt=stmt,
            length=r_l,
            start=r_s,
            end=r_e,
            txt=stmt.value,
            cur_idx=rng.cur_idx
        )


def get_selects(stmt):
    tokens = [x for x in stmt.tokens]
    for t in tokens:
        if t.is_group:
            tokens.extend(t.tokens)
        if (t.value == '(' and 
            [x for x in t.parent.tokens if x.ttype not in whitespace_tokens][1].value.lower().split(' ')[0] in select_node_words):
            yield t.parent



def get_sel_node(stmt:SqlStatement,txt:str) -> SelectNode:
        
    pos = stmt.cur_idx
    result = None
    length = len(stmt.txt)
    
    for sel in get_selects(stmt.stmt):
        s = stmt.txt.find(sel.value) + stmt.start
        e = s + len(sel.value)
        l = e-s
        
        # print(sel.value)
        # print(txt[s:e])
        # print(txt[pos-5:pos])
        if pos >= s and pos <= e and l <= length:
            result = sel
            length = l
            r_s, r_e, r_l = s,e,l
            
    if result:
        return SelectNode(
            stmt=sel,
            length=r_l,
            start=r_s,
            end=r_e,
            txt=result.value,
            cur_idx=stmt.cur_idx
        )
    else:
        return stmt


re_patterns = [
    # '(?:"""--sql[^\n]*\n)(?P<sql>.*?)(?:""")',
    # '([^\n]*"""--sql[^\n]*\n)(?P<sql>.*?)([\n]*?""")',
    '([^\n]*?"""--sql[^\n]*?\n)(?P<sql>.*?)(""")',
    # '([^\n]*?"""--sql[^\n]*?\n)(?P<sql>.*?)([\n]*?\s*?""")',
    '(.(sql|execute|executemany)\(")(?P<sql>.*?)("\s*(\)|,))',
    "(.(sql|execute|executemany)\(')(?P<sql>.*?)('\s*(\)|,))",
    ]


def get_idx(txt,cur_line,cur_col):
    x = sum(len(x)+1 for i,x in enumerate(txt.split('\n'),start=0) if i < cur_line)
    return x + cur_col

def get_range(txt,cur_line,cur_col):
    # print(len(txt))
    idx = get_idx(txt,cur_line,cur_col)
    # print(idx)
    re_matches = chain.from_iterable(
        (regex.finditer(
            pattern=p,
            string=txt,
            flags=regex.DOTALL | regex.IGNORECASE) for p in re_patterns)
    )
    
    for m in re_matches:
        s,e = m.span('sql')
        if idx >= s and idx <= e:
            # print(f'{s},{e}\n',m.groups()[0])
            txt = m.group('sql')
            if not txt or len(txt.strip()) == 0:
                return
            # if txt.strip()[-1] != ';':
            #     txt += ';'
            
            return SqlTxtRange(
                txt=txt,
                length=e-s,
                start=s,
                end=e,
                cur_idx=idx
            )


def get_sql2(txt,cur_line,cur_col) -> SelectNode:
    
    result = get_range(txt,cur_line,cur_col)
    return result

def get_sql(txt,cur_line,cur_col) -> SelectNode:
    
    result = get_range(txt,cur_line,cur_col)
    if not result:
        return
    
    result = get_statement(result,txt)
    if not result:
        return
    
    sel_node = get_sel_node(result,txt)
    
    if not sel_node:
        return result
    
    return sel_node


def left_of_cur_matches(line_txt:str,search_txt_list:list[str]):
    return any([line_txt.lower()[-len(search_txt):] == search_txt for search_txt in search_txt_list])


# table_predicates = ['from ','join ','pivot ','unpivot ']


# %%



def clean_partial_sql(txt):
    
    placeholder = (f'placeholder{x}' for x in range(2000))
    
    patterns = [
        ('(\w+[.])([\n\s)])',   f'\g<1>{next(placeholder)}\g<2>'),
        ('(\(\s*)(\))',         f'\g<1>{next(placeholder)}\g<2>'),
        ('(=[^\n\s]*$)',         f'\g<1> {next(placeholder)}'),
        ]
    
    for pat, rep in patterns:
        txt = regex.sub(pat,rep,txt,flags=regex.IGNORECASE)
    return txt





from pygments import highlight as _highlight
from pygments.lexers import SqlLexer
from pygments.formatters import HtmlFormatter


dataframe_types = {
    "pandas.core.frame.DataFrame":'pandas',
    "polars.dataframe.frame.DataFrame":'polars',
    "pyarrow.lib.Table":'arrow',
    "duckdb.DuckDBPyRelation":'duckdb_rel',
    # "duckdb.duckdb.DuckDBPyRelation":'duckdb_rel',
}

def check_dataframe_type(type_str:str):
    for k,v in dataframe_types.items():
        if k in type_str:
            return v
    


gui_style = """
QLabel {
	font-family: Consolas;
	font-size: 10pt;
}


QTreeWidget {
	font-family: Consolas;
	font-size: 10pt;
}
QTextEdit {
	font-family: Consolas;
	font-size: 10pt;
}
QListView {
	font-family: Consolas;
	font-size: 10pt;
}

QTableView {
	font-family: Consolas;
	font-size: 10pt;
}

QHeaderView{
	font-family: Consolas;
	border:0px;
	font-size: 10pt;
    font-weight: bold;
}

QTableView::item {
	padding: 0px;
}"""

format_map = {
        'STRING':'',
        'INT':',.0f',
        'NUMBER':',.2f',
        'DATE':'%x',
        'DATETIME':'%x',
        'TIME':'%X',
    }

fmt_types = {
    'BIGINT': 'INT',
    'BIT': 'STRING',
    'BOOLEAN': 'STRING',
    'BLOB': 'STRING',
    'DATE': 'DATE',
    'DOUBLE': 'NUMBER',
    'DECIMAL(s, p)': 'NUMBER',
    'HUGEINT': 'INT',
    'INTEGER': 'INT',
    'INTERVAL': 'DATE',
    'REAL': 'NUMBER',
    'SMALLINT': 'INT',
    'TIME': 'TIME',
    'TIMESTAMP': 'DATETIME',
    'TIMESTAMP WITH TIME ZONE': 'DATETIME',
    'TINYINT': 'INT',
    'UBIGINT': 'INT',
    'UINTEGER': 'INT',
    'USMALLINT': 'INT',
    'UTINYINT': 'INT',
    'UUID': 'STRING',
    'VARCHAR': 'STRING'
}


def highlight(text):
    # Generated HTML contains unnecessary newline at the end
    # before </pre> closing tag.
    # We need to remove that newline because it's screwing up
    # QTextEdit formatting and is being displayed
    # as a non-editable whitespace.
    highlighted_text = _highlight(text, SqlLexer(), HtmlFormatter()).strip()

    # Split generated HTML by last newline in it
    # argument 1 indicates that we only want to split the string
    # by one specified delimiter from the right.
    parts = highlighted_text.rsplit("\n", 1)

    # Glue back 2 split parts to get the HTML without last
    # unnecessary newline
    highlighted_text_no_last_newline = "".join(parts)
    return highlighted_text_no_last_newline
 

def apply_fmt(x,c,fmts):
    if x is None:
        return ''
    return f'{x:{fmts[c]}}'



import os
import re
from dataclasses import dataclass
from pathlib import Path
from enum import Enum
from lark import Lark, Transformer, v_args, Discard, Visitor, Tree, UnexpectedToken, Token, UnexpectedCharacters
from lark.parsers.lalr_interactive_parser import InteractiveParser
import duckdb
import logging
import time

lark_cache = Path(__file__).parent.joinpath('lark_cache')
lark_file = Path(__file__).parent.joinpath('sql3b.lark')

def parser_error_handler(e:UnexpectedToken):
    # assert isinstance(e, UnexpectedToken)

    
    if '_AS' in e.accepts and e.token.type == 'RPAREN':
        e.interactive_parser.feed_token(Token('_AS', 'AS'))
        e.interactive_parser.feed_token(Token('NAME', 'placeholder'))
        e.interactive_parser.feed_token(e.token)
        return True
    
    if e.token == Token('$END', '') and 'NAME' in e.accepts:
        e.interactive_parser.feed_token(Token('NAME', 'xyz'))
        return True
    
    if 'NAME' in e.accepts:
        e.interactive_parser.feed_token(Token('NAME', 'xyz'))
        e.interactive_parser.feed_token(e.token)
        return True
    
    # print(e.token)
    # e.interactive_parser.feed_eof() 
    # e.interactive_parser.feed_token(Token('NAME', 'xyz'))
    # e.interactive_parser.feed_token(e.token)
    return False


def get_parser():
    sql_grammer = lark_file.read_text()
    sql_parser = Lark(
        sql_grammer,
        parser="lalr",
        cache=str(lark_cache),
        propagate_positions=True,
        maybe_placeholders=True,
        debug=False,
    )
    return sql_parser

sql_parser = get_parser()

@dataclass
class Cte:
    name: str
    sql: str
    projection: list[str]
    cte_start: int
    self_start: int


@dataclass
class Ctes:
    sql: str
    map: dict[str, Cte]
    start_pos: int
    end_pos: int
    parent: "Query" = None


@dataclass
class Query:
    ctes: Ctes
    sql: str
    projection: list[str]
    from_refs: dict[str, 'FromRef']
    start_pos: int
    end_pos: int
    cte_sibblings: dict[str, Cte]
    set_operation:bool


class FromRefKind(Enum):
    table = 0
    table_function = 1
    subquery = 2


@dataclass
class FromRef:
    kind: FromRefKind
    alias: str
    name: str
    sql: str = None
    start_pos: int = None
    projection: list[tuple[str,str]] = None


class TransformFromClause(Transformer):
    def __init__(self, sql: str, parser:'SqlParserNew') -> None:
        super().__init__()
        self.sql = sql
        self.parser = parser
        self.map = {}

    def db_table(self, value):
        if value[1]:
            alias = value[1]
        else:
            alias = value[0]

        self.map[alias] = FromRef(kind=FromRefKind.table, alias=alias, name=value[0])
        # return {alias:FromRef(kind=FromRefKind.table,alias=alias,name=value[0])}

    def join_type(self, value):
        return Discard

    def table_name(self, value):
        tbl = ".".join([x.value for x in value if x])
        return tbl

    def alias(self, value):
        return value[0].value

    def table_function(self, value):
        if value[1]:
            
            sql = self.sql[value[0].meta.start_pos : value[0].meta.end_pos]
            
            self.map[value[1]] = FromRef(
                kind=FromRefKind.table_function,
                alias=value[1],
                name=value[0],
                sql=sql,
                projection=self.parser.db_describe_columns(f'select * from {sql}')
            )
            # self.sql[value[0].meta.start_pos:value[0].meta.end_pos]

    @v_args(tree=True)
    def subquery(self, tree: Tree):
        value = tree.children
        if value[1]:
            alias = value[1]
            sql = value[0]
            self.map[alias] = FromRef(
                kind=FromRefKind.subquery,
                alias=alias,
                name=sql,
                sql=sql,
                start_pos=tree.meta.start_pos+1,
            )
            # return {alias:sql}
        else:
            return Discard


class QueryToSql(Transformer):
   def __init__(self, sql: str) -> None:
      super().__init__()
      self.sql = sql

   @v_args(tree=True)
   def query(self, value: Tree):
      sql = self.sql[value.meta.start_pos : value.meta.end_pos]
      return sql


class GetQueries(Visitor):
   def __init__(self, sql: str, parser:'SqlParserNew',pos) -> None:
      super().__init__()
      self.sql = sql
      self.parser = parser
      self.queries: dict[int, Query] = {}
      self.queries_list: list[Query] = []
      self.q_start_end:list[tuple[int,int]] = []
      self.pos = pos

   def get_from_tables(self, tree: Tree):
      s1 = QueryToSql(self.sql).transform(tree)
      fr = TransformFromClause(self.sql,self.parser)
      fr.transform(s1).pretty()
      return fr.map

   def get_projections(self, tree: Tree):
      col_names:list[str] = []
      for col in tree.children:
         # print(col)
         if len(col.children) < 2:
            continue
         if col.children[1]:
            col_names.append(col.children[1].children[0].value)
            continue
         if col.children[0].data == 'col_ref':
            col_names.append(col.children[0].children[1].value)
            continue
      return [(x,None) for x in col_names]

   def get_ctes(self, cte: Tree):
      if not cte:
         return
      if cte.data != "cte":
        #  print(cte.pretty())
         raise Exception("not cte")

      cte_map = {}

      for x in cte.children:
         if not isinstance(x, Tree):
               continue
         name = x.children[0].value

         q = self.get_query(x.children[3])
         if not q:
            continue    

         c = Cte(
               name=name,
               sql=self.sql[x.children[3].meta.start_pos : x.children[3].meta.end_pos],
               projection=q.projection,
               cte_start=cte.meta.start_pos,
               self_start=x.children[3].meta.start_pos,
         )

         cte_map[name] = c

      cte_sql_txt = self.sql[cte.meta.start_pos : cte.meta.end_pos]

      return Ctes(
         sql=cte_sql_txt,
         map=cte_map,
         start_pos=cte.meta.start_pos,
         end_pos=cte.meta.end_pos,
      )

   def get_query(self, tree: Tree):
      if not tree:
        return
      cte_data = None
      set_op = False
      cols = []
      if tree.children[0]:
         if tree.children[0].data == "query" and tree.data == 'query':
            q = self.get_query(tree.children[0])
            return q
         if tree.data == "set_operation":
            set_op = True
            q1 = self.get_query(tree.children[0])
            cols = q1.projection
            
      
         if tree.children[0].data == "cte":
            cte_data = self.get_ctes(tree.children[0])
            # print(tree.pretty())
            # raise Exception("not cte")
      # print(cte_data)
      sql_txt = self.sql[tree.meta.start_pos : tree.meta.end_pos]
      from_refs = {}
      for x in tree.children:
         # print(x)
         if not isinstance(x, Tree):
               continue
         if x.data == "from_clause":
               # print(2,x.pretty())
               from_refs = self.get_from_tables(x)
               # print(2,from_refs)
         if x.data == "select_clause":
            cols = self.get_projections(x.children[4])
      start = tree.meta.start_pos
      end = tree.meta.end_pos
      q = Query(
         ctes=cte_data,
         sql=sql_txt,
         projection=cols,
         from_refs=from_refs,
         start_pos=start,
         end_pos=end,
         cte_sibblings={},
         set_operation=set_op,
      )
      
      return q
   
   def set_operation(self, tree: Tree):
      self.query(tree)
      
      
   def query(self, tree: Tree):

      q = self.get_query(tree)
      if not q:
         return
      if (q.start_pos,q.end_pos) in self.q_start_end:
         return
      self.q_start_end.append((q.start_pos,q.end_pos))
      self.queries_list.append(q)
      self.queries[q.start_pos] = q


incomplete_col_ref = re.compile('(\w+[.])([\n\s),])')

class SqlParserNew:
    
    def __init__(self, db: duckdb.DuckDBPyConnection = None, ls = None, logger:logging.Logger = None, file_search_path:str = None) -> None:
        self.db = db
        self.projection_cache = {}
        self.ls = ls
        self.file_search_path = file_search_path
        self.log = logger.getChild('sql_parser')
        self.log_describe = logger.getChild('sql_describe')
        self.log_query_output = logger.getChild('query_output')
        self.log_interactive_parser = logger.getChild('interactive_parser')
        

    def show_message(self,msg):
        if self.ls:
            self.ls.show_message_log(msg)



    def parse_sql(self, sql: str, pos:int):

        
        try:
            tree, choices_pos = interactive_parse(sql,pos,self.log_interactive_parser)
            # self.log.debug(['interactive_parse',tree,choices_pos])
            # tree = sql_parser.parse(sql,on_error=parser_error_handler)
        except UnexpectedToken as e:
            self.log.info(['failed to parse, Unexpected Token',sql,e,e.token])
            return None, None
        except Exception as e:
            self.log.info(['failed to parse',sql,e])
            return None, None
            # noqa: E722
        if not tree:
            return None, choices_pos
        queries = GetQueries(sql,self,pos)
        queries.visit(tree)


        #add cte sibblings
        subq_lu = {}
        cte_lu = {}
        for q in queries.queries_list:
            if not q.ctes:
                continue
            for k, v in q.ctes.map.items():
                cte_lu[v.self_start] = v.cte_start
            for k, v in q.from_refs.items():
                if v.kind == FromRefKind.subquery:
                    subq_lu[v.start_pos] = q.start_pos

        for q in queries.queries_list:
            if q.start_pos not in cte_lu:
                continue
            parent_map = queries.queries[cte_lu[q.start_pos]].ctes.map
            sibblings = {k: v for k, v in parent_map.items() if v.self_start < q.start_pos}
            q.cte_sibblings = sibblings

        for q in queries.queries_list:
            if q.start_pos not in subq_lu:
                continue
            q.cte_sibblings = queries.queries[subq_lu[q.start_pos]].ctes.map

        for q in queries.queries_list:
            if q.set_operation:
                for q_in_set in [x for x in queries.queries_list if x.start_pos >= q.start_pos and x.end_pos <= q.end_pos]:
                    q_in_set.cte_sibblings = q.cte_sibblings
        
        
        
        if not self.db:
            # self.show_message(f'no db connection')
            return queries
        
        # self.show_message(f'running db queries')
        
        # cte_and_subq = [cte for k,cte in q.ctes.map.items() for q in queries.queries_list if q.ctes]
        # [subq for k,subq in q.from_refs.items() for q in queries.queries_list if subq.kind == FromRefKind.subquery] 
        
        for q in queries.queries_list:
            cte_sql = ''
            if q.cte_sibblings:
                sibblings = ',\n'.join([f'{k} as ({v.sql})' for k,v in q.cte_sibblings.items()])
                cte_sql = f'with {sibblings}'
            sql = f'{cte_sql}\n{q.sql}'
            if not q.end_pos:              #testing
                q.end_pos = len(sql)       #testing

            if pos >= q.start_pos and pos <= q.end_pos:
                continue
            projection = self.db_describe_columns(sql)
            if projection:
                q.projection = projection
        
        
        for q in queries.queries_list:
            if q.ctes:
                for k,v in q.ctes.map.items():
                    if v.self_start in queries.queries:
                        v.projection = queries.queries[v.self_start].projection
        
        
        # self.log_query_output.debug(queries.queries_list)
        return queries, choices_pos
    

    def db_describe_columns(self, sql):
        if sql in self.projection_cache:
            return self.projection_cache[sql]
        
        if incomplete_col_ref.search(f'{sql} '):
            # self.log.debug(['skipping describe col, incomplete col ref',sql])
            return

        try:
            db = self.db.cursor()
            if self.file_search_path and len(self.file_search_path)>0:
                db.execute(f"set file_search_path = '{self.file_search_path}';")
            rec = db.execute(f'describe ({sql})').fetchall()
            data = [(x[0],x[1]) for x in rec]
            self.projection_cache[sql] = data
            return data
        except Exception as e:  # noqa: E722
            # self.log.debug(['failed to run describe',sql,e,os.getcwd()])
            return
        


def find_end(p:InteractiveParser,cur_token:Token=None):
    choices = list(p.choices().keys())
    if 'table_ref' in choices and cur_token.upper() in ['FROM','JOIN']:
        try:
            p.feed_token(Token('IDENT', 'placeholder'))
            choices = list(p.choices().keys())
        except:
            pass


    if '$END' in choices:
        try:
            return p.feed_eof()
        except Exception as e:
            pass


@dataclass
class TokenHistory:
    token:Token
    choices:list
    accept:list

@dataclass
class ParseResult:
    parser:InteractiveParser
    tree:Tree
    choices:list[str]
    token_history:list[TokenHistory]
    tokens_to_pos:list[TokenHistory]
    duration:float


no_space_tokens = set([".", ",", ";", "(", ")", "[", "]","{","}"])



def interactive_parse(sql:str,pos:int,logger:logging.Logger, parser=sql_parser):
    start = time.time()
    p = parser.parse_interactive(sql)
    tokens = p.iter_parse()
    token_history:list[TokenHistory] = []

    token_history.append(
        TokenHistory(
            token=None,
            choices=list(p.choices().keys()),
            accept=list(p.accepts())
        )
    )
    # tk = next(lex)
    choices_pos = []
    if logger:
        # logger.debug(['interactive_parse',sql,pos])
        pass
    
    token = None
        
    while True:
        prev_token = token
        try:
            token = next(tokens)
        except StopIteration:
            break
        except UnexpectedCharacters as e:
            # print(e)
            break
        except UnexpectedToken as e:
            # print(e)
            break
        
        token_history.append(
            TokenHistory(
                token=prev_token,
                choices=list(p.choices().keys()),
                accept=list([])
            )
        )

    token_history.append(
        TokenHistory(
            token=prev_token,
            choices=list(p.choices().keys()),
            accept=list([])
        )
    )

    result_history:list[TokenHistory] = []

    for i, t in enumerate(token_history):
        if t.token is None:
            result_history.append(t)
            continue
        if t.token.end_pos < pos or (t.token.end_pos <= pos and t.token in no_space_tokens):
            result_history.append(t)
        else:
            break
    t = result_history[-1]
    tree=find_end(p,cur_token=token)
    choices_pos = t.choices
    # print(t)
    return tree, choices_pos


#%%

def line_col(str:str, idx:int):
    return str.count('\n', 0, idx), idx - str.rfind('\n', 0, idx)-1

def get_idx(txt,cur_line,cur_col):
    x = sum(len(x)+1 for i,x in enumerate(txt.split('\n'),start=0) if i < cur_line)
    return x + cur_col

def move_line(txt:str, idx:int, anchor:int, direction:int):
    row, col = line_col(txt, idx)
    arow, acol = line_col(txt, anchor)
    start = min(row,arow)
    end = max(row,arow)
    lines = txt.split('\n')
    print(start,end,idx,anchor)
    if direction == -1:
        if min(row,arow) == 0:
            return None
        else:
            new_txt = '\n'.join(lines[0:start - 1] + lines[start:end+1] + [lines[start-1]] + lines[end+1:])
            new_idx = get_idx(new_txt, row-1, col)
            new_anchor = get_idx(new_txt, arow-1, acol)
            return new_idx, new_anchor, new_txt
    if direction == 1:
        if max(row,arow) + 1 == len(lines):
            return None
        else:
            new_txt = '\n'.join(lines[0:start] + [lines[end+1]] + lines[start:end+1]  + lines[end+2:])
            new_idx = get_idx(new_txt, row+1, col)
            new_anchor = get_idx(new_txt, arow+1, acol)
            return new_idx, new_anchor, new_txt
            

# %%



############################################################################
# Copyright(c) Open Law Library. All rights reserved.                      #
# See ThirdPartyNotices.txt in the project root for additional notices.    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################
import sys
import asyncio
from typing import Optional
import sqlglot
import time

from lsprotocol import types as lsp
from lsprotocol.types import (
    CompletionList,
)

from dabbler.lsp.parser import sql_parser
from lark import UnexpectedToken

from dabbler.lsp.server_classes import InlineSqlLangServer
from dabbler.lsp.sql_utils import (
    get_sql2,
    line_col,
    get_range,
    get_statement,
)
from dabbler.lsp.completion import pathlib_completetions


if sys.platform == 'win32':
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
sql_server = InlineSqlLangServer("pygls-json-example", "v0.1")


# list("abcdefghijklmnopqrstuvwxyz")+


# trigger_characters=[":",'.',' '],
@sql_server.thread()
@sql_server.feature(
    lsp.TEXT_DOCUMENT_COMPLETION,
    lsp.CompletionOptions(
        trigger_characters=[":", ".",'/', ' ','"',"'"],
    ),  # all_commit_characters=[":"]
)
def completions(
    ls: InlineSqlLangServer, params: Optional[lsp.CompletionParams] = None
) -> CompletionList:
    """Returns completion items."""
    # ls.zmq_check_for_update()
    # ls.check_sockets()
    document = ls.workspace.get_document(params.text_document.uri)
    current_line_txt = document.lines[params.position.line]


    trigger = params.context.trigger_character

    # sql_rng = get_sql(document.source, params.position.line, params.position.character)
    sql_rng = get_sql2(document.source, params.position.line, params.position.character)

    # ls.show_message_log(f'{params}')
    # ls.show_message_log(f'\nsql_rng: {sql_rng.start,sql_rng.end}\nleft of idx{document.source[sql_rng.cur_idx-5:sql_rng.cur_idx]}\n{sql_rng.txt}\n{document.source[sql_rng.start:sql_rng.end]}')


    if not sql_rng:
        # ls.show_message_log(f"txt {current_line_txt[:params.position.character]}")
        try:
            path_completions = pathlib_completetions(current_line_txt[:params.position.character], ls.pathlibs_paths, ls.log)
            if path_completions:
                return path_completions
        except Exception as e:
            ls.log.debug(['pathlib completion error',e])
        return None

    if trigger in ['"',"'"]:
        return None

    pos_in_range = sql_rng.cur_idx - sql_rng.start

    if not ls.completer:
        return None

    # try:
    comps = ls.completer.route_completion2(
        pos_in_range, sql_rng, trigger, params.position.line, current_line_txt
    )
    # ls.show_message_log(f'completion time: {time.time()-start}')

    # except:
    # ls.show_message_log('problem with completion')
    # return None
    return comps




def publish_diagnostics(ls:InlineSqlLangServer, uri, line, char,  msg):
    d = lsp.Diagnostic(
            range=lsp.Range(
                start=lsp.Position(line=line, character=char),
                end=lsp.Position(line=line, character=char+1),
            ),
            message=msg,
            source='inline_sql',
        )
    ls.publish_diagnostics(uri,[d])


@sql_server.thread()
@sql_server.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: InlineSqlLangServer, params: lsp.DidChangeTextDocumentParams):
    """Text document did change notification."""
    if not params.content_changes[0].range:
        return
    # ls.show_message_log(f'did change: {params}')
    # ls.show_message_log(f'did change: {params.content_changes[0].range.start}')
    start_line = params.content_changes[0].range.end.line
    char = params.content_changes[0].range.end.character
    document = ls.workspace.get_document(params.text_document.uri)
    # current_line_txt = document.lines[start_line]
    # ls.show_message_log(f"did change: {start_line,char}, line txt: {current_line_txt}")
    sql_range = get_sql2(document.source, start_line, char)
    if not sql_range:
        # ls.show_message_log(f"did not find sql range at line:{start_line + 1}")
        ls.publish_diagnostics(params.text_document.uri, [])
        return

    try:
        sql_range.txt = sql_range.txt.replace('\r\n','\n')
        # ls.show_message_log(f"sql: {sql_range}")
        start = time.time()
        p = sql_parser.parse(sql_range.txt)
        # ls.log.debug(['parse time',time.time()-start])


    except Exception as e:
        # ls.log.debug(['parse time',time.time()-start])
        rng_start_line, rng_start_col = line_col(
        ls.workspace.get_document(params.text_document.uri).source, sql_range.start
        )
        
        if isinstance(e, UnexpectedToken):
            msg = f'Unexpected Token "{e.token}"'
            # ls.show_message_log(f'line: {e.line}, col: {e.column} {msg}')
            publish_diagnostics(ls, params.text_document.uri, rng_start_line+e.line-2, e.column, msg)
            
    
    else:
        ls.publish_diagnostics(params.text_document.uri, [])



@sql_server.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
async def did_open(ls: InlineSqlLangServer, params: lsp.DidOpenTextDocumentParams):
    ls.log_workspace_info()
    ls.save_key_file()
    ls.create_default_compelter()


@sql_server.command(InlineSqlLangServer.CMD_SEND_SQL_TO_GUI)
def send_sql_to_gui(ls: InlineSqlLangServer, *args):
    """Starts counting down and showing message asynchronously.
    It won't `block` the main thread, which can be tested by trying to show
    completion items.
    """
    q = None
    try:
        uri_data = args[0][0]
        pos = args[0][1]
    except:
        ls.show_message_log("send cmd gui: error getting uri and pos")
        return

    if "line" not in pos or "character" not in pos:
        ls.show_message_log("did not find line/charager in pos")
        return

    if "external" not in uri_data:
        ls.show_message_log("problem obtaining document uri")
        return

    uri = uri_data["external"]
    line = pos["line"]
    char = pos["character"]

    document = ls.workspace.get_document(uri)
    sql_rng = get_sql2(document.source, line, char)

    if not sql_rng:
        ls.show_message_log(
            f"did not find sql range at line:{line + 1} char: {char + 1}"
        )

    pos_in_range = sql_rng.cur_idx - sql_rng.start
    try:
        q, queries, choices_pos = ls.completer.get_queries(pos_in_range,sql_rng.txt)
    except Exception as e:
        ls.log.debug(['error getting queries',e])

    if not q:
        ls.show_message_log("did not find query")
        return
    
    cte_sql = ''
    if q.cte_sibblings:
        sibblings = ',\n'.join([f'{k} as ({v.sql})' for k,v in q.cte_sibblings.items()])
        cte_sql = f'with {sibblings}'
    sql = f'{cte_sql}\n{q.sql}'
    
    msg = {"cmd": "run_sql", "sql": sql}

    resp = ls.zmq_send(msg)

    # ls.show_message_log(f"args: {args}\n")
    return


# @sql_server.feature(lsp.TEXT_DOCUMENT_FORMATTING)
@sql_server.command(InlineSqlLangServer.CMD_FORMAT_CURRENT_STATEMENT)
def format_range(ls: InlineSqlLangServer, *args):
    """Starts counting down and showing message asynchronously.
    It won't `block` the main thread, which can be tested by trying to show
    completion items.
    """
    try:
        uri_data = args[0][0]
        pos = args[0][1]
    except:
        ls.show_message_log("send cmd gui: error getting uri and pos")
        return

    if "line" not in pos or "character" not in pos:
        ls.show_message_log("did not find line/charager in pos")
        return

    if "external" not in uri_data:
        ls.show_message_log("problem obtaining document uri")
        return

    uri = uri_data["external"]
    line = pos["line"]
    char = pos["character"]

    document = ls.workspace.get_document(uri)
    sql_rng = get_range(document.source, line, char)
    if not sql_rng:
        return

    sql_rng = get_statement(sql_rng, document.source)
    if not sql_rng:
        return

    try:
        p = sqlglot.parse_one(sql_rng.txt)
    except:
        ls.show_message_log("problem parsing sql")
        return

    sql = p.sql(pretty=True, dialect="duckdb", pad=4, indent=4)
    sql = "\n".join([f"    {x}" for i, x in enumerate(sql.split("\n"))]) + "\n"
    s_line, s_char = line_col(document.source, sql_rng.start)
    e_line, e_char = line_col(document.source, sql_rng.end)
    rng = lsp.Range(lsp.Position(s_line - 1, 0), lsp.Position(e_line - 1, 0))

    text_edit = lsp.TextEdit(rng, sql)

    # ls.send_notification('textDocument/didChange', {
    # 'textDocument': {
    #     'uri': document.uri,
    #     'version': document.version
    # },
    # 'contentChanges': [{
    #     'range': rng,
    #     'rangeLength': len(text_edit.range),
    #     'text': text_edit.new_text
    # }]
    # })
    ls.apply_edit(lsp.WorkspaceEdit(changes={document.uri: [text_edit]}))
    # ls.show_message_log(f"args: {args}\n")
    return [text_edit]


# validation_error = []

# def get_validations():
#     return validation_error

# @json_server.feature(
#     lsp.TEXT_DOCUMENT_DIAGNOSTIC,
#     lsp.DiagnosticOptions(
#         identifier="jsonServer",
#         inter_file_dependencies=False,
#         workspace_diagnostics=False,
#     ),
# )
# def text_document_diagnostic(ls:InlineSqlLangServer,
#     params: lsp.DocumentDiagnosticParams,
# ) -> lsp.DocumentDiagnosticReport:
#     """Returns diagnostic report."""
#     document = json_server.workspace.get_document(params.text_document.uri)
#     ls.show_message(f"val: {validation_error}")
#     return lsp.RelatedFullDocumentDiagnosticReport(
#         items=get_validations(),
#         kind=lsp.DocumentDiagnosticReportKind.Full,
#     )

# @json_server.feature(lsp.WORKSPACE_DIAGNOSTIC)
# def workspace_diagnostic(ls,
#     params: lsp.WorkspaceDiagnosticParams,
# ) -> lsp.WorkspaceDiagnosticReport:
#     """Returns diagnostic report."""
#     first = list(json_server.workspace._docs.keys())[0]
#     document = json_server.workspace.get_document(first)
#     ls.show_message(f"val: {validation_error}")
#     return lsp.WorkspaceDiagnosticReport(
#         items=[
#             lsp.WorkspaceFullDocumentDiagnosticReport(
#                 uri=document.uri,
#                 items=validation_error,
#                 kind=lsp.DocumentDiagnosticReportKind.Full,
#             )
#         ]
#     )


def main():
    import logging
    from pathlib import Path

    # logging.basicConfig(filename=Path(__file__).parent.joinpath("pygls.log"), filemode="w", level=logging.DEBUG)
    sql_server.start_io()
    # sql_server.check_sockets2()
    # sql_server.create_sockets2()


if __name__ == "__main__":
    main()



from filecmp import cmp
from math import log
from lsprotocol.types import (
    CompletionItemKind,
    MarkupContent,
)
from pathlib import Path
from dabbler.lsp.sql_utils import CmpItem
import duckdb
import json
from dabbler.common import check_name
import logging
logger = logging.getLogger(__name__)


get_records_sql = """--sql
    with
        cols as (
            select 
                database_name||'.'||schema_name as db_scm,
                table_name, 
                list([column_name, data_type]) as cols
            from duckdb_columns()
            group by all
        ),
    tables_views as (
        select 
            database_name||'.'||schema_name as db_scm,
            table_name, 
            'table' as object_type,
            'table' as table_type,
            case
                when not internal then sql
                else null
            end as sql
        from duckdb_tables()
        union all
        select 
            database_name||'.'||schema_name as db_scm,
            view_name, 
            'table' as object_type,
            'view' as table_type,
            case
                when not internal then sql
                else 'internal'
            end as sql
        from duckdb_views()
        union all
            select distinct
            database_name||'.'||schema_name as db_scm,
            d.function_name,
            'table' as object_type,
            'table_macro',
            case
                when not internal then 'external'
                else 'internal'
            end as sql
        from duckdb_functions() d
        where d.function_type in ('table','table_macro')
            --and d.function_name not in ('force_checkpoint')
        union all
        select distinct
            database_name||'.'||schema_name as db_scm,
            d.function_name,
            'function' as object_type,
            d.function_type,
            null,
        from duckdb_functions() d
        where d.function_type in ('aggregate','macro','scalar')
            and d.function_name not similar to '.*\W.*'
            and not d.internal
            and database_name != 'system'
        group by all
        union all
        select distinct
            null,
            d.function_name,
            'function' as object_type,
            d.function_type,
            'internal',
        from duckdb_functions() d
        where d.function_type in ('aggregate','macro','scalar')
            and d.function_name not similar to '.*\W.*'
            and database_name = 'system'
        group by all
    )
    select 
        t.*,
        case when object_type = 'table' then c.cols end as cols
    from tables_views t
        left join cols c using (db_scm, table_name)
    """




def make_db(db_data:dict):
    """
    makes a duckdb in memory database from mirroring the tables 
    and columns in the IPython database.
    
    This DB is used by the language server to get column completion items
    for subqueries and CTEs.
    """
    db2 = duckdb.connect()
    databases = [x[0] for x in db2.execute("select database_name from duckdb_databases()").fetchall()]

    for db_to_add in db_data['databases']:
        if db_to_add not in databases:
            try:
                db2.execute(f"attach ':memory:' as {db_to_add}")
            except Exception as e:
                logger.error(f'Error attaching dummy memory database {db_to_add}, {e}')

    schemas = [x[0] for x in db2.execute("select database_name ||'.'|| schema_name from duckdb_schemas()").fetchall()]

    for schema in db_data['schemas']:
        if schema not in schemas:
            try:
                db2.execute(f"create schema {schema}")
            except Exception as e:
                logger.error(f'Error creating schema {schema}, {e}')
            
    db2.execute(f"use {db_data['current_schema']};")

    for df in db_data['dataframes']:
        try:
            db2.execute(df[1])
        except Exception as e:
            logger.error(f'Error creating table for dataframe {df[0]}, {e}')


    for schema, item, sql, cols in db_data['data']:
        if not cols:
            continue
        col_names = []
        col_items = []
        for c, d in cols:
            if c in col_names:
                continue
            col_names.append(c)
            col_items.append([c, d])

        col_txt = ',\n'.join([f'"{c[0]}" {c[1]}' for c in col_items])
        sql2 = f'create table {check_name(item)}({col_txt})'
        try:
            db2.execute(f'use {schema}; {sql2}')
        except Exception as e:
            logger.error(f'Error creating table {item} in schema {schema}, {e}')

    db2.execute(f"use {db_data['current_schema']};")
    if db_data['file_search_path']:
        try:
            db2.execute(f"""set file_search_path to '{db_data['file_search_path']}';""")
        except Exception as e:
            logger.error(f'Error setting file_search_path to {db_data["file_search_path"]}, {e}')
    
    return db2



def make_completion_map(db:duckdb.DuckDBPyConnection,db_data):
    """makes a map of completion items for the language server"""
    records = db.execute(get_records_sql).fetchall()
    function_docs = json.loads(Path(__file__).parent.joinpath('functions.json').read_text())
    kind_map = {
        'table':CompletionItemKind.File,
        'table_macro':CompletionItemKind.File,
        'function':CompletionItemKind.Function,
    }

    sort_map = {
        'table':"2",
        'function':"8"
    }


    item_map:dict[str,list[CmpItem]] = {}
    item_map['root_namespace'] = []

    cur_db = db_data['current_schema'].split('.')[0]

    for db_scm, fn_name, fn_type in db_data['functions']:
        if fn_type == 'table_macro':
            obj_type = 'table_macro'
        else:
            obj_type = 'function'

        comp_item = CmpItem(
                label=fn_name,
                kind=kind_map[obj_type],
                detail=None,
                typ='function',
                sort='9',
                obj_type=obj_type,
                doc=None)
        if db_scm not in item_map:
            item_map[db_scm] = []
        item_map[db_scm].append(comp_item)


    for db_scm, item, obj_type, comp_detial, sql, cols in records:

        
        item = check_name(item)

        if not db_scm:
            db_scm = 'root_namespace'

        db_schema_split = db_scm.split('.')
        if len(db_schema_split) == 2:
            db, schema = db_schema_split
        else:
            db = 'aa'
            schema = 'aabb'

        if db_scm not in item_map:
            item_map[db_scm] = []

        if db == cur_db:
            if schema not in item_map:
                item_map[schema] = []
        
        fn_doc = None
        fn_detail = None
        
        sort = sort_map[obj_type]
        if sql == 'internal':
            sort = '9'
        
        if obj_type == 'function' or comp_detial == 'table_macro':
            if item in function_docs:
                fn_doc = MarkupContent(
                    kind='markdown',
                    value=function_docs[item]['documentation']['documentation']) 
                fn_detail = function_docs[item]['documentation']['detail'] 
        
        comp_item = CmpItem(
                label=item,
                kind=kind_map[obj_type],
                detail=fn_detail,
                typ=comp_detial,
                sort=sort,
                obj_type=obj_type,
                doc=fn_doc)
        

        item_map[db_scm].append(comp_item)
        if db == cur_db:
            item_map[schema].append(comp_item)

        if db_scm == db_data['current_schema']:
            item_map['root_namespace'].append(comp_item)
        
        
        if cols:
            col_completions = [CmpItem(
                label=c[0],
                kind=CompletionItemKind.Field,
                detail=None,
                typ=c[1],
                sort='1',
                obj_type='column',
                doc=None) for c in cols]
            item_map[f'{db_scm}.{item}'] = col_completions
            if db == cur_db:
                item_map[f'{schema}.{item}'] = col_completions
            if db_scm == db_data['current_schema']:
                if item in item_map:
                    item_map[item].extend(col_completions)
                else:
                    item_map[item] = col_completions



    for cat_schema in db_data['schemas']:
        cat, schema = cat_schema.split('.')
        cat_comp = CmpItem(
            label=cat,
            kind=CompletionItemKind.Folder,
            detail=None,
            typ='database',
            sort='3',
            obj_type='database',
            doc=None)
        
        schema_comp = CmpItem(
            label=schema,
            kind=CompletionItemKind.Folder,
            detail=None,
            typ='schema',
            sort='3',
            obj_type='schema',
            doc=None)
        
        if cat == cur_db:
            if schema not in item_map['root_namespace']:
                # if not cat_schema in item_map:
                    # item_map[cat_schema] = []
                # item_map[schema] = item_map[cat_schema]
                item_map['root_namespace'].append(schema_comp)

        if cat not in item_map:
            item_map[cat] = []
            item_map['root_namespace'].append(cat_comp)
            
        item_map[cat].append(schema_comp)
        
        root_labels = [x.label for x in item_map['root_namespace']]
        item_map['root_namespace'].extend([x for x in item_map['system.main'] if x.label not in root_labels])
    
    # for cat_schema in db_data['schemas']:
    #     cat, schema = cat_schema.split('.')
    #     if cat == cur_db:
    #         if schema not in item_map and schema not in item_map['root_namespace']:
    #             if not cat_schema in item_map:
    #                 item_map[cat_schema] = []
    #             item_map[schema] = item_map[cat_schema]
    #             item_map['root_namespace'].append(schema_comp)
    
    return item_map


        
   


#%%
from ast import In
import logging
import os
import re
from dataclasses import dataclass
from pathlib import Path
from enum import Enum
from lark import Lark, Transformer, v_args, Discard, Visitor, Tree, UnexpectedToken, Token, UnexpectedCharacters, UnexpectedEOF, UnexpectedInput
from lark.parsers.lalr_interactive_parser import InteractiveParser
import duckdb
import time

lark_cache = Path(__file__).parent.joinpath('lark_cache')
lark_file = Path(__file__).parent.joinpath('sql3b.lark')

def get_parser():
    sql_grammer = lark_file.read_text()
    sql_parser = Lark(
        sql_grammer,
        parser="lalr",
        cache=str(lark_cache),
        propagate_positions=True,
        maybe_placeholders=True,
        debug=False,
    )
    return sql_parser

sql_parser = get_parser()



check_choices = (
    ('RPAREN', ')'),
    ('NAME', 'placeholder'),
)


def find_end(p:InteractiveParser,cur_token:Token=None):
    choices = list(p.choices().keys())
    print(cur_token,choices)
    if 'table_ref' in choices and cur_token.upper() in ['FROM','JOIN']:
        try:
            p.feed_token(Token('IDENT', 'placeholder'))
            choices = list(p.choices().keys())
        except:
            pass


    if '$END' in choices:
        try:
            return p.feed_eof()
        except Exception as e:
            print(e)
    # for typ, value in check_choices:
    #     if typ in choices:
    #         t = Token(typ, value)
    #         print(f'feeding {t}')
    #         p.feed_token(t)
    #         return find_end(p)

@dataclass
class TokenHistory:
    token:Token
    choices:list
    accept:list

@dataclass
class ParseResult:
    parser:InteractiveParser
    tree:Tree
    choices:list[str]
    token_history:list[TokenHistory]
    tokens_to_pos:list[TokenHistory]
    duration:float


no_space_tokens = set([".", ",", ";", "(", ")", "[", "]","{","}"])


def interactive_parse_new(sql:str,pos:int,logger:logging.Logger = None):
    start = time.time()
    p = sql_parser.parse_interactive(sql)
    tokens = p.iter_parse()
    token_history:list[TokenHistory] = []

    token_history.append(
        TokenHistory(
            token=None,
            choices=list(p.choices().keys()),
            accept=list(p.accepts())
        )
    )
    # tk = next(lex)
    choices_pos = []
    if logger:
        logger.debug(['interactive_parse',sql,pos])
    
    token = None
        
    while True:
        prev_token = token
        try:
            token = next(tokens)
        except StopIteration:
            break
        except UnexpectedCharacters as e:
            print(e)
            break
        except UnexpectedToken as e:
            print(e)
            break
        
        token_history.append(
            TokenHistory(
                token=prev_token,
                choices=list(p.choices().keys()),
                accept=list([])
            )
        )

    token_history.append(
        TokenHistory(
            token=prev_token,
            choices=list(p.choices().keys()),
            accept=list([])
        )
    )

    result_history:list[TokenHistory] = []

    for i, t in enumerate(token_history):
        if t.token is None:
            result_history.append(t)
            continue
        if t.token.end_pos < pos or (t.token.end_pos <= pos and t.token in no_space_tokens):
            result_history.append(t)
        else:
            break
    t = result_history[-1]
    tree=find_end(p,cur_token=token)
    choices_pos = t.choices
    # print(t)
    return ParseResult(
        parser=p,
        tree=tree,
        choices=choices_pos,
        token_history=[f'{x.token} {x.token.start_pos}:{x.token.end_pos}' for x in token_history if x.token],
        tokens_to_pos=[f'{x.token} {x.token.start_pos}:{x.token.end_pos}' for x in result_history if x.token],
        duration=time.time()-start,
    )


    

    # if (len(token_history) > 0
    #     and token_history[-1]
    #     and token_history[-1].lower() == 'from'
    #     and 'IDENT' in p.accepts()
    #     and 'table_ref' in p.choices()):
    #     p.feed_token(Token('IDENT', 'placeholder'))

    # if not choices_pos:
    #     choices_pos = list(p.choices().keys())
    # tree = find_end(p,cur_token=token)
    # if logger:
    #     logger.debug({'choices_pos':choices_pos})
    #     logger.debug({'tree':tree})
    # return tree, choices_pos, token_history

#%%




#%%
import duckdb
import os
from IPython import get_ipython
from dabbler.gui_stuff import check_dataframe_type
from dabbler.common import check_name
from pathlib import Path


def get_db_data_new(db:duckdb.DuckDBPyConnection, file_search_path:str=None):
    """gets the data to send to the language server"""
    db_items = db.execute("""--sql
            with
        cols as (
            select 
                database_name||'.'||schema_name as db_scm,
                table_name, 
                list([column_name, case when starts_with(data_type,'EMUM') then 'ENUM(''DUMMY'')' else data_type end]) as cols
            from duckdb_columns
            group by all
        ), db_data as (
            (select database_name||'.'||schema_name as db_scm, table_name, sql from duckdb_tables)
            union all
            (select database_name||'.'||schema_name, view_name, sql from duckdb_views order by view_oid)
            union all
            (select database_name||'.'||schema_name, sequence_name, sql from duckdb_sequences())
        )
        select
            d.*,
            c.cols
        from db_data d
            left join cols c using (db_scm, table_name)
        """).fetchall()

    db_functions = db.sql(
            """--sql,
            select
                database_name||'.'||schema_name as db_scm,
                function_name,
                function_type,
            from duckdb_functions()
            WHERE internal = false
            """
            ).fetchall()

    taken_table_names = [x[0] for x in db.sql(
        """
        select distinct 
            table_name 
        from duckdb_columns() 
        where 
            database_name = current_database()
            and schema_name = current_schema()
            
            union all 
        
        select distinct 
            function_name 
        from duckdb_functions() 
        where 
            function_type = 'table'
            and database_name = current_database()
            and schema_name = current_schema()
        """).fetchall()]

    paths = []
    dataframes = []
    ipython = get_ipython()
    for item in (ipython.ev('dir()')):
        i_type = str(type(ipython.ev(item)))
        df_type = check_dataframe_type(i_type)

        if 'WindowsPath' in i_type or 'PosixPath' in i_type:
            paths.append(
                (item, str(ipython.ev(item)))
            )

        if i_type == "<class 'module'>":
            if '__file__' not in ipython.ev(f'dir({item})'):
                continue
            if 'site-packages' in ipython.ev(item).__file__:
                continue
            if Path(ipython.ev(item).__file__).parent.name == 'Lib':
                continue
            if Path(ipython.ev(item).__file__).parent.parent.name == 'Lib':
                continue
            
            for item2 in ipython.ev(f'dir({item})'):
                item_item2 = f'{item}.{item2}'
                i_type2 = str(type(ipython.ev(item_item2)))

                if 'WindowsPath' in i_type2 or 'PosixPath' in i_type2:
                    paths.append(
                        (item_item2, str(ipython.ev(item_item2)))
                    )
        
        if df_type and item[0] != '_' and item not in taken_table_names:
            unique_name = f'my_item_zz_{item}'
            locals().__setitem__(unique_name,ipython.ev(item))
            try:
                cols = [[x[0],x[1]] for x in  db.sql(f"describe select * from {unique_name} limit 1").fetchall()]
            except:
                continue
            
            sql = f'CREATE TABLE {item}({", ".join([f"{check_name(c[0])} {c[1]}" for c in cols])});'
            
            dataframes.append([
                item,
                sql,
                cols])

    paths.sort(key=lambda x: f'{x[0]}-{x[1]}')
    current_schema:str = db.execute("select current_database()||'.'||current_schema()").fetchone()[0]
    databases = [x[0] for x in db.execute("select database_name from duckdb_databases()").fetchall()]
    schemas = [x[0] for x in db.execute("select database_name ||'.'|| schema_name from duckdb_schemas()").fetchall()]

    db_data = {'data':db_items,
            'dataframes':dataframes,
            'databases':databases,
            'functions':db_functions,
            'paths':paths,
            'schemas':schemas,
            'current_schema':current_schema,
            'cwd':os.getcwd(),
            'file_search_path':file_search_path,
            }

    return db_data










# def get_db_data(db:duckdb.DuckDBPyConnection):
#     records = db.execute(
#         """
#         select
#             c.table_catalog,
#             c.table_schema,
#             c.table_name,
#             t.table_type,
#             list([column_name, data_type]),
#             sql
#         from duckdb_columns c
#             join information_schema.tables t using (table_catalog, table_schema, table_name)
#         group by all
        
#         union all
        
#         select distinct
#             d.database_name,
#             d.schema_name,
#             d.function_name,
#             'table_function',
#             [],
#             null
#         from duckdb_functions() d
#         where d.function_type in ('table','table_macro')
#             --and d.function_name not in ('force_checkpoint')
#         group by all
        
        
#         """).fetchall()
    
#     functions = db.execute(
#         """
#         select distinct
#             d.database_name,
#             d.schema_name,
#             d.function_name,
#             d.function_type,
#             [],
#         from duckdb_functions() d
#         where d.function_type in ('aggregate','macro','scalar')
#             and d.function_name not similar to '.*\W.*'
#             and not d.internal
#             and database_name != 'system'
#         group by all
        
#         union all
        
#         select distinct
#             null,
#             null,
#             d.function_name,
#             d.function_type,
#             [],
#         from duckdb_functions() d
#         where d.function_type in ('aggregate','macro','scalar')
#             and d.function_name not similar to '.*\W.*'
#             and database_name = 'system'
#         group by all
#         """).fetchall()

#     table_names = [x[2] for x in records]

#     ipython = get_ipython()
#     for item in (ipython.ev('dir()')):
#         i_type = str(type(ipython.ev(item)))
        
#         if i_type in capture_types and item[0] != '_' and item not in table_names:
#             locals().__setitem__(item,ipython.ev(item))
#             cols = [[x[0],x[1]] for x in  db.sql("describe select * from df limit 1").fetchall()]
#             records.append([None,None,item,'python',cols])

#     current_schema = db.execute("select current_schema()").fetchone()[0]
#     current_db = db.execute("select current_database()").fetchone()[0]
#     sequences = db.execute("select database_name, schema_name, sequence_name, from duckdb_sequences()").fetchall()
    
#     db_data = {
#         'tbl_col':records,
#         'sequences':sequences,
#         'functions':functions,
#         'current_db':current_db,
#         'current_schema':current_schema,
#     }
    
#     return db_data






#%%



from IPython.core.magic import cell_magic, magics_class, Magics
from queue import Queue
from dabbler.gui_main import MyApp


@magics_class
class DbDabbler(Magics):
    def __init__(self, ipython, debug=False):
        super(DbDabbler, self).__init__(ipython)
        db = None
        ev_dir = ipython.ev("dir()")
        if '__file__' in ev_dir:
            file = ipython.ev('__file__')
        else:
            file = ipython.ev('__vsc_ipynb_file__')
        for item in ipython.ev("dir()"):
            if "duckdb.DuckDBPyConnection" in str(type(ipython.ev(item))):
                db_name = item
                db = ipython.ev(item)
                file_search_path = db.execute(
                    "select current_setting('file_search_path')"
                ).fetchone()[0]
                db = db.cursor()
                if file_search_path:
                    db.execute(f"set file_search_path to '{file_search_path}'")
                break
        if db:
            self.q = None
            self.app = MyApp([], db, debug=debug, file_search_path=file_search_path, file=file, db_name=db_name)
            ipython.run_line_magic("gui", "qt")
            from IPython.lib.guisupport import start_event_loop_qt4

            start_event_loop_qt4(self.app)
        else:
            print("db not found, create duckdb db before loading extension 'db = duckdb.connect()'")

    @cell_magic
    def runq(self, line, cell):
        if self.q:
            start_txt = '"""--sql\n'
            start = cell.find(start_txt) + len(start_txt)
            end_txt = '"""'
            end = cell[start:].find(end_txt) + start
            sql = cell[start:end]
            print(sql)
            self.q.put(sql)


from dabbler.ext_classes import DbDabbler


def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.

    # app = .instance()
    # if app is None:

    magics = DbDabbler(ipython)
    ipython.register_magics(magics)



from pathlib import Path
import tempfile
import json
import uuid
import duckdb
from typing import Literal, TypedDict, Union
import logging
from logging.handlers import SocketHandler
import pprint
import pickle
import struct
import re
from dabbler.lsp.parser import lark_file

def get_grammer_kw():
    grammer_txt = lark_file.read_text()
    reg = re.compile(r'''([A-Z_]+)\s*:\s*"[A-Z_]+"i''')
    return reg.findall(grammer_txt)


duckdb_keyworkds = set(x[0].upper() for x in duckdb.execute("select keyword_name from duckdb_keywords() where keyword_category = 'reserved'").fetchall())
# duckdb_keyworkds_all = set(x[0].upper() for x in duckdb.execute("select keyword_name from duckdb_keywords()").fetchall())
duckdb_keyworkds.update([
    'DATE',
    ])


grammer_kw = get_grammer_kw()



def check_name(col:str):
    if col:
        if (' ' in col or
            '-' in col or
            '/' in col or
            '.' in col or
            col.upper() in duckdb_keyworkds):
            return f'"{col}"'
        return col



class PprintSocketHandler(SocketHandler):
    def makePickle(self, record):
        """
        Pickles the record in binary format with a length prefix, and
        returns it ready for transmission across the socket.
        """
        ei = record.exc_info
        if ei:
            # just to get traceback text into record.exc_text ...
            dummy = self.format(record)
        # See issue #14436: If msg or args are objects, they may not be
        # available on the receiving end. So we convert the msg % args
        # to a string, save it as msg and zap the args.
        d = dict(record.__dict__)
        # d['msg'] = record.getMessage()
        d['msg'] = pprint.pformat(record.msg)
        d['args'] = None
        d['exc_info'] = None
        # Issue #25685: delete 'message' if present: redundant with 'msg'
        d.pop('message', None)
        s = pickle.dumps(d, 1)
        slen = struct.pack(">L", len(s))
        return slen + s

class FromLangServer(TypedDict):
    cmd: Literal['run_sql', 'db_data_update', 'check_for_update','connection_id','heartbeat']
    data: Union[str,dict,int]
    con_id: int
    

class ToLangServer(TypedDict):
    cmd: Literal['db_data','ip','no_update','connection_id','ip_python_started','debug','heartbeat']
    data: Union[str,dict,int,bool]
    con_id: int
    
class ConnInfo(TypedDict):
    workspace_path: str
    main_port: int
    handshake_port: int
    server_id: int
    client_id: int


class Connections(TypedDict):
    connections: dict[str,ConnInfo]


class KeyFile:
    def __init__(self) -> None:
        self.file = Path(tempfile.gettempdir()) / 'dabbler_key.json'
        if self.file.exists():
            self.read()
        else:
            self.connections = {}
            self.save()
    
    def save(self):
        self.file.write_text(json.dumps(self.connections))
    
    def read(self):
        self.connections = json.loads(self.file.read_text())
        
    def add_connection(self, name: str, conn_info: ConnInfo):
        conn_info['server_id'] = uuid.uuid4().int
        conn_info['client_id'] = uuid.uuid4().int
        self.connections[name] = conn_info
        self.save()
        return conn_info

    def delete_connection(self, name: str):
        del self.connections[name]
        self.save()
        
    def get_connection(self, file: Path) -> ConnInfo:
        workspaces = [Path(x) for x in self.connections if Path(x).exists() and Path(x) in file.parents]
        if len(workspaces) == 0:
            return None
        workspaces.sort(key=lambda x: file.parents.index(x))
        workspace = workspaces[0]
        return self.connections[str(workspace)]

    


#%%

from qtpy import QtWidgets, QtGui, QtCore
from txt_util import line_col, move_line, get_idx
from gui_stuff import gui_style
from dabbler.lsp.new_parser import interactive_parse_new


key_map = {

}

class Shortcut:
    def __init__(self, Seq, Parent, target, description:str=None):
        pp = f'{Seq: <10}{description}'
        key_map[Seq] = pp
        if not isinstance(Seq, list):
            Seq = [Seq]
        for i in Seq:
            sc = QtGui.QShortcut(QtGui.QKeySequence(i), Parent)
            sc.activated.connect(target)


class ShortcutHelp(QtWidgets.QWidget):
    def __init__(self, parent:'SqlEditor'):
        super().__init__()
        self.setWindowTitle('Shortcuts')
        # self.setGeometry(100, 100, 800, 600)
        self.app2 = parent

        self.text_edit = QtWidgets.QLabel()
        self.text_edit.setText('\n'.join(key_map.values()))

        #make always on top
        self.setWindowFlags(QtCore.Qt.WindowType.WindowStaysOnTopHint)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.text_edit)
        self.setLayout(self.layout)
        self.short_keys = [ 
            Shortcut("Ctrl+H", self, self.app2.show_shortcuts, 'Show shortcuts'),
            Shortcut("Ctrl+w", self, self.app2.show_shortcuts, 'close window'),
            Shortcut("Ctrl+q", self, self.app2.show_shortcuts, 'close window'),
        ]

class SqlTextBox(QtWidgets.QPlainTextEdit):
    def __init__(self,parent:'EditorWindow'=None):
        super().__init__()    
        # self.document().contentsChanged.connect(self.txt_change)
        # self.textChanged.connect(self.txt_change)
        self.editor = parent
        self.cursorPositionChanged.connect(self.cursor_move)

    def cursor_move(self):
        cursor = self.textCursor()
        pos = cursor.position()
        self.editor.cur_pos.setText(f'cursor_pos: {pos}')

        result = interactive_parse_new(self.toPlainText(), pos)
        if result:
            if result.token_history:
                self.editor.all_toekns_box.clear()
                self.editor.all_toekns_box.addItems(result.token_history)
            else:
                self.editor.all_toekns_box.clear()

            self.editor.tree_text.clear()
            if result.tree:
                self.editor.tree_text.setPlainText(result.tree.pretty())

            if result.choices:
                self.editor.choices_box.clear()
                self.editor.choices_box.addItems(result.choices)
            else:
                self.editor.choices_box.clear()

            if result.tokens_to_pos:
                self.editor.token_history.clear()
                self.editor.token_history.addItems(result.tokens_to_pos)
            else:                
                self.editor.token_history.clear()
            if result.duration:
                self.editor.duartion.addItem(str(result.duration))

        else:
            self.editor.choices_box.clear()
            self.editor.token_history.clear()
            self.editor.all_toekns_box.clear()


    def keyPressEvent (self, keyEvent:QtGui.QKeyEvent):
        vis = self.editor.comp.popup().isVisible()
        # print(keyEvent.key(), keyEvent.nativeModifiers(),f'popup: {vis}')
        if keyEvent.key() == 16777218 and keyEvent.nativeModifiers() == 513: #shift tab key
                        # print('tab key')
            keyEvent = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                32, #space key
                QtCore.Qt.KeyboardModifiers(keyEvent.nativeModifiers())
            )
            txt = self.toPlainText()
            idx = self.textCursor().position()
            anchor = self.textCursor().anchor()
            row, col = line_col(txt, idx)
            arow, acol = line_col(txt, anchor)
            lines = txt.split('\n')
            start = min(row,arow)
            end = max(row,arow)
            for i in range(start, end+1):
                if lines[i].startswith('    '):
                    lines[i] = lines[i][4:]
                else:
                    lines[i] = lines[i].lstrip()

            new_txt = '\n'.join(lines)
            self.setPlainText(new_txt)
            cursor = self.textCursor()
            cursor.setPosition(get_idx(new_txt, row, len(lines[row])-1), QtGui.QTextCursor.MoveMode.MoveAnchor)
            cursor.setPosition(get_idx(new_txt, arow, len(lines[arow])-1), QtGui.QTextCursor.MoveMode.KeepAnchor)
            self.setTextCursor(cursor)
            return

        if keyEvent.key() == 16777217 and keyEvent.nativeModifiers() == 512: #tab key
            # print('tab key')
            keyEvent = QtGui.QKeyEvent(
                QtCore.QEvent.Type.KeyPress,
                32, #space key
                QtCore.Qt.KeyboardModifiers(keyEvent.nativeModifiers())
            )
            txt = self.toPlainText()
            idx = self.textCursor().position()
            anchor = self.textCursor().anchor()
            row, col = line_col(txt, idx)
            arow, acol = line_col(txt, anchor)
            lines = txt.split('\n')
            start = min(row,arow)
            end = max(row,arow)
            for i in range(start, end+1):
                lines[i] = '    ' + lines[i]

            new_txt = '\n'.join(lines)
            self.setPlainText(new_txt)
            cursor = self.textCursor()
            cursor.setPosition(get_idx(new_txt, row, col+4), QtGui.QTextCursor.MoveMode.MoveAnchor)
            cursor.setPosition(get_idx(new_txt, arow, acol+4), QtGui.QTextCursor.MoveMode.KeepAnchor)

            self.setTextCursor(cursor)
            return
        super().keyPressEvent(keyEvent)


class EditorWindow(QtWidgets.QWidget):
    def __init__(self, parent:'SqlEditor'):
        super().__init__()
        self.setWindowTitle('SQL Editor')
        self.setGeometry(100, 100, 1600, 900)
        self.app2 = parent

        self.text_edit = SqlTextBox(self)
        self.text_edit.setPlaceholderText('Enter your SQL here')
        # self.text_edit.setFont(QtGui.QFont('Arial', 12))


        # self.monaco_widget.setLanguage('sql')
        self.token_history = QtWidgets.QListWidget()
        self.token_history.setFixedWidth(200)
        self.choices_box = QtWidgets.QListWidget()
        self.choices_box.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.MultiSelection)
        self.choices_box.setFixedWidth(200)
        self.all_toekns_box = QtWidgets.QListWidget()
        self.all_toekns_box.setFixedWidth(200)
        self.duartion = QtWidgets.QListWidget()
        self.duartion.setFixedWidth(100)

        self.cur_pos = QtWidgets.QLabel('cursor_pos: 0')
        self.header_label = QtWidgets.QLabel('sql editor')

        self.tree_text = QtWidgets.QPlainTextEdit()
        self.tree_text.setReadOnly(True)

        self.layoutz = QtWidgets.QGridLayout()
        self.layoutz.addWidget(self.cur_pos,0,0,QtCore.Qt.AlignmentFlag.AlignRight)
        self.layoutz.addWidget(self.text_edit, 1, 0, 2, 1)
        self.layoutz.addWidget(QtWidgets.QLabel('choices'),0,1)
        self.layoutz.addWidget(self.choices_box,1,1)
        self.layoutz.addWidget(QtWidgets.QLabel('choices'),2,1)
        self.layoutz.addWidget(QtWidgets.QLabel('all tokens'),0,2)
        self.layoutz.addWidget(self.all_toekns_box,1,2)
        self.layoutz.addWidget(QtWidgets.QLabel('tokens before cursor'),0,3)
        self.layoutz.addWidget(self.token_history,1,3)
        self.layoutz.addWidget(QtWidgets.QLabel('duration'),0,4)
        self.layoutz.addWidget(self.duartion,1,4)

        self.layoutz.addWidget(self.tree_text,1,5)


        self.setLayout(self.layoutz)

        
        self.short_keys = [
            Shortcut("Ctrl+H", self, self.show_shortcuts, 'show/hide shortcuts'),
            # Shortcut("Alt+Up", self, lambda:self.move(-1), 'Move line up'),
            # Shortcut("Alt+Down", self, lambda:self.move(1), 'Move line down'),
            Shortcut("Ctrl+N", self, self.app2.new_window, 'New window'),
            Shortcut("Ctrl+=", self, self.txt_bigger, 'zoom in'),
            Shortcut("Ctrl+-", self, self.txt_smaller, 'zoom out'),
            Shortcut("Ctrl+w", self, self.close, 'close window'),
            Shortcut("Ctrl+q", self, self.close, 'close window'),
            Shortcut("Ctrl+t", self, self.test_index, 'test'),

        ]
        self.text_edit.setPlainText('select\n   *\nfrom potato p')
        self.comp = QtWidgets.QCompleter(['apple', 'banana', 'cherry','lime'] + [f'word_{x}' for x in range(50)], self.text_edit)
        self.comp.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
        self.comp.setWidget(self.text_edit)
        self.comp_popup = self.comp.popup()
        for i in range(4):
            self.txt_bigger()

 

    def test_index(self):
        pos = self.text_edit.textCursor().position()
        qt_row = self.text_edit.textCursor().blockNumber()
        qt_col = self.text_edit.textCursor().columnNumber()
        char_w = self.text_edit.fontMetrics().width('a')
        char_h = self.text_edit.fontMetrics().height()
        rect = self.text_edit.cursorRect()
        

        self.comp.complete(QtCore.QRect((qt_col) * char_w + 3, (qt_row+1) * char_h + 8,  char_w * 40, 1))
        # if self.comp.popup().isVisible():
        #     self.comp.popup().hide()
        # else:
        #     self.comp.popup().show()
        txt = self.text_edit.toPlainText()
        line, col = line_col(txt, pos)
        idx = get_idx(txt, line, col)
        print(f'pos: {pos}, rect: {rect}, line: {line}, col: {col}, idx: {idx} qtrow: {qt_row}, qtcol{qt_col}, {char_w}, {char_h}')

    def show_shortcuts(self):
        self.app2.show_shortcuts()
        self.activateWindow()

    # def text_selected(self):
    #     cursor = self.text_edit.textCursor()
    #     txt = cursor.selectionStart(), cursor.selectionEnd()
    #     cursor.setPosition(4, QtGui.QTextCursor.MoveMode.MoveAnchor)
    #     cursor.setPosition(8, QtGui.QTextCursor.MoveMode.KeepAnchor)
    #     cursor.anchor()
    #     cursor.position()
    #     self.text_edit.setTextCursor(cursor)
    #     print(f'text selected {txt}')

    def txt_bigger(self):
        editor_font = self.text_edit.font()
        popup_font = self.comp_popup.font()
        new_size = editor_font.pointSize() + 1
        editor_font.setPointSize(new_size)
        popup_font.setPointSize(new_size)
        self.text_edit.setFont(editor_font)
        self.comp_popup.setFont(popup_font)
        self.choices_box.setFont(editor_font)
        self.token_history.setFont(editor_font)
        self.all_toekns_box.setFont(editor_font)
        
    
    def txt_smaller(self):
        editor_font = self.text_edit.font()
        popup_font = self.comp_popup.font()
        new_size = editor_font.pointSize() - 1
        editor_font.setPointSize(new_size)
        popup_font.setPointSize(new_size)
        self.text_edit.setFont(editor_font)
        self.comp_popup.setFont(popup_font)
        self.choices_box.setFont(editor_font)
        self.token_history.setFont(editor_font)
        self.all_toekns_box.setFont(editor_font)

    def tst_move(self):
        cursor = self.text_edit.textCursor()
        cursor.setPosition(4, QtGui.QTextCursor.MoveMode.MoveAnchor)
        self.text_edit.setTextCursor(cursor)

    def move(self,direction:int):
        txt = self.text_edit.toPlainText()
        pos = self.text_edit.textCursor().position()
        anchor = self.text_edit.textCursor().anchor()
        result = move_line(txt, pos, anchor, direction)
        if result:
            new_idx, newanchor, new_txt = result
            self.text_edit.setPlainText(new_txt)
            cursor = self.text_edit.textCursor()
            cursor.setPosition(newanchor, QtGui.QTextCursor.MoveMode.MoveAnchor)
            cursor.setPosition(new_idx, QtGui.QTextCursor.MoveMode.KeepAnchor)
            self.text_edit.setTextCursor(cursor)

    def closeEvent(self, a0: None) -> None:
        if len(self.app2.windows) == 1:
            self.app2.quit()
        return super().closeEvent(a0)



class SqlEditor(QtWidgets.QApplication):
    def __init__(self):
        super().__init__([])
        self.windows:list[EditorWindow] = []
        self.new_window()
        self.shortcut_keys = [
        ]
        self.popup = ShortcutHelp(self)

        
        
    def new_window(self):
        window = EditorWindow(self)
        window.show()

        

        self.windows.append(window)

    def run(self):
        self.setStyleSheet(gui_style)
        self.exec_()

    def quit(self):
        super().quit()


    def show_shortcuts(self):
        if self.popup.isVisible():
            self.popup.hide()
        else:
            self.popup.show()


app = SqlEditor()
app.run()

# %%



